import pytest
from conftest import btl_path, cnv_path, hex_path

from ctdam.parser import (
    BottleFile,
    CnvCollection,
    CnvFile,
    DataFile,
    FileCollection,
    HexCollection,
    XMLCONFile,
    get_collection,
)


@pytest.mark.parametrize(
    ("path", "suffix", "file_type", "num_of_files"),
    [
        (btl_path, ".btl", BottleFile, 1),
    ],
)
class TestCollections:
    @pytest.fixture
    def files(
        self,
        path: str,
        suffix: str,
        file_type: DataFile,
        num_of_files: int,
    ):
        files = get_collection(
            path_to_files=path,
            file_suffix=suffix,
        )
        return files

    @pytest.fixture
    def df(self, files: FileCollection):
        return files.get_collection_dataframe()

    def test_collected_dataframe(self, files: FileCollection):
        assert (
            files.df.size
            == sum([file.df.shape[0] for file in files])
            * files.df_list[0].shape[1]
        )

    def test_cast_specific_info_added(
        self,
        files: FileCollection,
        num_of_files: int,
    ):
        df = files.get_collection_dataframe(
            files.get_dataframes(event_log=True, coordinates=True)
        )

        assert len(df.Event.unique()) == num_of_files
        assert len(df.Longitude.unique()) == num_of_files
        assert len(df.Latitude.unique()) == num_of_files

    def test_cnv_specifics(self, files: FileCollection):
        if files.file_suffix != "cnv":
            pytest.skip("cnv-specific test")
        assert files.data_meta_info["c0mS/cm"]["name"] == "Conductivity"
        df = files.tidy_collection_dataframe(files.df)
        assert df.isna().sum().sum() == 0
        assert files.array.shape == (
            sum([file.parameters.full_data_array.shape[0] for file in files]),
            files[0].parameters.full_data_array.shape[1],
        )
        assert files.array.shape == files.df.shape
        assert files.get_dataframes(time_correction=True)


def test_mixed_processing_steps():
    cnvs = CnvCollection(cnv_path, only_metadata=True)
    with pytest.warns(UserWarning):
        cnvs.get_processing_steps()


class TestHexCollections:
    @pytest.fixture
    def files(self) -> FileCollection:
        files = HexCollection(
            path_to_files=hex_path,
            pattern="SO308-2",
            sorting_key=lambda file: int(file.stem.split("_")[1]),
        )
        return files

    def test_pattern_handling(self, files):
        assert len(files) == 3

    def test_xmlcon_matching(self, files):
        xmlcon = XMLCONFile(hex_path.joinpath("SO308-2.XMLCON"))
        for file in files:
            assert file.xmlcon == xmlcon



import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join("..", "src")))

project = "ctdam"
copyright = "2023-2026, Emil Michels"
author = "Emil Michels"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",
    "sphinx.ext.todo",
    "sphinx.ext.linkcode",
    "sphinx_copybutton",
]


def linkcode_resolve(domain, info):
    if domain != "py":
        return None
    if not info["module"]:
        return None
    filename = info["module"].replace(".", "/")
    return (
        "https://github.com/DAM-CTD-Software/ctdam/src/branch/main/src/%s.py"
        % filename
    )


html_logo = "images/ctd_rosette.svg"
html_favicon = "images/ctd_rosette.svg"

autodoc_member_order = "bysource"
templates_path = ["_templates"]
exclude_patterns = []

html_theme = "furo"
html_static_path = ["_static"]



import importlib
import logging
from datetime import datetime, timezone
from pathlib import Path

import numpy as np

from ctdam.parser import BottleLogFile, CnvFile, DataFile
from ctdam.parser.ctddata import CTDData
from ctdam.proc.utils import BinnedDataError

logger = logging.getLogger(__name__)


class InvalidArgumentCombination(Exception):
    """Exception raised when an invalid combination of arguments is provided."""

    pass


class OwnBtlFile(DataFile):
    def __init__(
        self,
        ctd_data: CTDData | None = None,
        blf: BottleLogFile | None = None,
        path_to_file: Path | str = "",
    ):
        if ctd_data and blf:
            self.ctd_data = ctd_data
            self.blf = blf
            self.data = self.create_btl()
        elif path_to_file:
            # TODO: use DataFile variables and methods to read an existing file
            pass
        else:
            raise InvalidArgumentCombination

    def create_btl(self) -> str:
        req_parameters = [
            "prDM",
            "t090C",
            "t190C",
            "c0mS/cm",
            "c1mS/cm",
            "sbox0Mm/Kg",
            "sbox1Mm/Kg",
            "sal00",
            "sal11",
            "par",
            "spar",
            "flECO-AFL",
            "turbWETntu0",
        ]

        data_averages = self._get_averages()
        ctd_data_header = self.ctd_data.create_header()
        timestamp = datetime.now(timezone.utc).strftime("%Y.%m.%d %H:%M:%S")
        ctd_data_header.insert(
            -2,
            f"# create_bottlefile_metainfo = {timestamp}, ctdam python package, v{importlib.metadata.version('ctdam')}\n",
        )
        btl_file = "".join(
            [
                line
                for line in ctd_data_header
                if not line.startswith(
                    ("# name", "# span", "# nquan", "# nvalues", "# units")
                )
            ]
        )

        btl_base_id = int(self.ctd_data.metadata["WsStartID"]) - 1

        line = add_whitespace("Btl_Posn")
        line += add_whitespace("Btl_ID")
        line += add_whitespace("Datetime", 22)
        for i in range(len(req_parameters)):
            line += add_whitespace(req_parameters[i])
        btl_file += line + "\n"

        for i in range(len(self.blf.data_list)):
            line = ""

            line += add_whitespace(self.blf.data_list[i][0][1])
            line += add_whitespace(self.blf.data_list[i][0][1] + btl_base_id)
            line += add_whitespace(self.blf.data_list[i][1], 22)

            for j in range(len(data_averages[0])):
                line += add_whitespace(data_averages[i, j])

            if i == len(self.blf.data_list) - 1:
                btl_file += line
                continue

            btl_file += line + "\n"

        return btl_file

    def _get_averages(self) -> np.ndarray:
        cnv_data = self.ctd_data.parameters.get_full_data_array().astype(float)
        par_index_list = self._get_parameters()
        averages = np.array(
            [
                [None for x in range(len(par_index_list))]
                for y in range(len(self.blf.data_list))
            ]
        )

        for i in range(len(self.blf.data_list)):
            for j in range(len(par_index_list)):
                start_index = self.blf.data_list[i][2][0]
                end_index = self.blf.data_list[i][2][1]
                averages[i, j] = np.average(
                    cnv_data[start_index:end_index, j]
                ).round(4)

        return averages

    def _get_parameters(self) -> list:
        """gets the Indices of the required parameterts from the cnv

        Parameters
        ----------
        List of Parameters of the cnv

        Returns
        -------
        list of corresponding indices

        """
        req_parameters = [
            "prDM",
            "t090C",
            "t190C",
            "c0mS/cm",
            "c1mS/cm",
            "sbox0Mm/Kg",
            "sbox1Mm/Kg",
            "sal00",
            "sal11",
            "par",
            "spar",
            "flECO-AFL",
            "turbWETntu0",
        ]
        parameter_list = self.ctd_data.get_parameter_list()
        par_shortname_list = [x.name for x in parameter_list]
        par_index_list = []
        for i in range(len(req_parameters)):
            par_index_list.append(par_shortname_list.index(req_parameters[i]))
        return par_index_list


def _check_input(input, type):
    if type == BottleLogFile:
        if isinstance(input, BottleLogFile):
            return input
        elif isinstance(input, str):
            if input and Path(input).exists():
                return type(input)
            else:
                return ""
        elif isinstance(input, Path):
            if input is not Path(".") and input.exists():
                return type(input)
            else:
                return ""
        else:
            raise ValueError(
                f"Argument of {type(input)} cannot be used for {type}"
            )
    elif type == CTDData:
        if isinstance(input, CTDData):
            return input
        elif isinstance(input, str):
            if input:
                return CnvFile(input).to_ctd_data()
            else:
                return ""
        elif isinstance(input, Path):
            if input is not Path("."):
                return CnvFile(input).to_ctd_data()
            else:
                return ""
        else:
            raise ValueError(
                f"Argument of {type(input)} cannot be used for {type}"
            )
    else:
        raise ValueError(
            f"Argument of {type(input)} cannot be used for {type}"
        )


def create_bottle_file(
    input: CTDData | Path | str = "",
    arguments: dict = {},
    output_name: Path | str = "",
    original_input_path: Path | str = "",
    **kwargs,
) -> OwnBtlFile | CTDData:
    """
    Creates a custom bottle file, given a .cnv and .bl file.

    The resulting file strongly adheres to the format of a regular .btl file.
    Specifically, the header is the same, only the data table features a
    different format. Its a 11-character wide tsv, as a cnv data table. In
    contrast to a .btl, only average values are used.

    In general, this custom bottle file (.obtl) can be generated at any time
    during the CTD processing. This improves over the standard Sea-Bird variant
    that allows this only during .cnv creation using Datcnv. With the .obtl
    file one can ensure the very same data quality from a .cnv file inside a
    bottle file.
    """
    ctd_data = _check_input(input, CTDData)
    try:
        blf = _check_input(arguments["bl"], BottleLogFile)
    except KeyError:
        if original_input_path:
            original_input_path = Path(original_input_path)
        else:
            original_input_path = ctd_data.path_to_file
        if original_input_path.exists():
            blf = _check_input(
                original_input_path.with_suffix(".bl"), BottleLogFile
            )
        else:
            blf = None

    if ctd_data:
        if not blf:
            try:
                blf = BottleLogFile(ctd_data.path_to_file.with_suffix(".bl"))
            except (FileNotFoundError, ValueError, TypeError):
                logger.info(
                    f"Could not find a corresponding .bl file to the cnv {ctd_data.metadata_source.path_to_file}"
                )
                return ctd_data
    else:
        if blf:
            try:
                ctd_data = CnvFile(
                    blf.path_to_file.with_suffix(".cnv")
                ).to_ctd_data()
            except (FileNotFoundError, ValueError, TypeError):
                raise ValueError(
                    f"Could not find a corresponding .cnv file to the bl {blf.path_to_file}"
                )
        else:
            raise InvalidArgumentCombination

    if ctd_data.binned:
        raise BinnedDataError(
            file_name=ctd_data.file_name,
            step_name="create_bottle_file",
        )
    btl = OwnBtlFile(ctd_data, blf)

    # usually write btl to disk, skip this only when explicitely stated
    if "write_btl" in arguments and not arguments["write_btl"]:
        pass
    else:
        if not output_name:
            output_name = ctd_data.path_to_file
        if "file_suffix" in arguments:
            stem = Path(output_name).stem
            output_name = Path(output_name).with_stem(
                stem + arguments["file_suffix"]
            )
        with open(Path(output_name).with_suffix(".obtl"), "w") as file:
            file.write(btl.data)

    return btl


def add_whitespace(data, space: int = 11):
    return (space - len(str(data))) * " " + str(data)



import copy
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from conftest import cnv_path
from numpy.testing import assert_equal

from ctdam.parser import CnvFile, CTDData
from ctdam.proc.module import MissingParameterError
from ctdam.proc.modules import AlignCTD
from ctdam.proc.modules.geomar_wildedit import WildeditGEOMAR
from ctdam.proc.modules.seabird_functions import (
    BinAvg,
    CellTM,
    LoopRemoval,
    WFilter,
)
from ctdam.proc.procedure import Procedure
from ctdam.proc.utils import (
    BinnedDataError,
    get_alignment_delay_and_correlation_values,
)


@pytest.mark.parametrize(
    "file",
    [file for file in cnv_path.glob("*.cnv")],
)
class TestExampleFiles:
    @pytest.fixture
    def cnv(self, file: Path, run_long_tests: bool) -> CnvFile:
        if (not run_long_tests) and (file.stat().st_size > 5 * 1000000):
            pytest.skip(f"Skipping long cnv: {file}, due to pytest option.")
        try:
            cnv = CnvFile(file)
        except FileNotFoundError as error:
            pytest.skip(str(error))
        else:
            return cnv

    @pytest.fixture
    def df(self, cnv: CnvFile) -> pd.DataFrame:
        return cnv.df

    def test_align_function(self, cnv: CnvFile, create_files):
        instance = AlignCTD()
        try:
            result = instance(
                input=cnv,
                arguments={
                    "Oxygen": None,
                },
            )
        except (ValueError, BinnedDataError):
            assert True
        else:
            # check, whether the returned delay value is below a threshold
            if create_files:
                result.to_cnv(
                    f"alignctd_{result.metadata_source.file_name}.cnv"
                )
            if cnv.processing_steps.get_step("alignctd"):
                delay_values = get_alignment_delay_and_correlation_values(
                    result.processing_steps
                )
                for delays in delay_values:
                    assert 1 < float(delays[0]) < 6
            else:
                assert True

    def test_wildedit_geomar(self, cnv: CnvFile, create_files):
        instance = WildeditGEOMAR()
        try:
            return_cnv = instance(
                input=cnv,
                arguments={
                    "std1": 3,
                    "std2": 10,
                    "window_size": 200,
                },
            )
        except (ValueError, BinnedDataError):
            assert True
        else:
            if create_files:
                return_cnv.to_cnv(
                    f"wildedit_{return_cnv.metadata_source.file_name}.cnv"
                )
            assert isinstance(return_cnv, CTDData)
            assert isinstance(
                return_cnv.parameters.full_data_array, np.ndarray
            )

    def test_procedure_on_cnvs(self, cnv: CnvFile):
        old_cnv = copy.deepcopy(cnv)
        proc_config = {
            "input": cnv,
            "output_type": "internal",
            "modules": {
                "wildedit_geomar": {"window_size": 200},
                "alignctd": {
                    "Oxygen": None,
                    "Conductivity": None,
                    "Salinity": None,
                },
            },
        }
        try:
            Procedure(
                proc_config,
                timeout=5,
            )
        except (MissingParameterError, BinnedDataError):
            assert True
        else:
            # handle case, when no columns were present inside the file
            if (
                "alignctd" or "wildedit_geomar"
            ) in cnv.processing_steps.get_names():
                assert len(cnv.processing_steps) > len(
                    old_cnv.processing_steps
                )
            else:
                assert True

    def test_w_filter(self, cnv, create_files):
        instance = WFilter()
        pre_cnv = copy.deepcopy(cnv.parameters)
        try:
            return_cnv = instance(
                input=cnv,
                arguments={
                    "Temperature": {"window_width": 50},
                    "Oxygen": {"half_width": 2, "offset": 5},
                    "Conductivity": {
                        "window_type": "triangle",
                        "window_width": 200,
                        "half_width": 3,
                        "offset": 3,
                    },
                },
            )
        except (ValueError, BinnedDataError):
            assert True
        else:
            assert isinstance(return_cnv, CTDData)
            for param in return_cnv.parameters.get_parameter_list():
                if param.param not in [
                    "Pressure",
                    "Temperature",
                    "Conductivity",
                    "Salinity",
                    "Oxygen",
                    "Fluorometer",
                    "Turbidity Meter",
                    "PAR",
                    "SPAR",
                ]:
                    continue
                try:
                    if create_files:
                        return_cnv.to_cnv(
                            f"wfilter_{return_cnv.metadata_source.file_name}.cnv"
                        )
                    assert_equal(pre_cnv[param.name].data, param.data)
                except AssertionError:
                    assert True
                else:
                    print(f"Param still equal: {param.name}")
                    assert False

    def test_cell_tm(self, cnv, create_files):
        instance = CellTM()
        pre_cnv = copy.deepcopy(cnv)
        try:
            return_cnv = instance(
                input=cnv,
                arguments={},
            )
        except (ValueError, BinnedDataError):
            assert True
        else:
            assert isinstance(return_cnv, CTDData)
        for param in return_cnv.parameters.get_parameter_list():
            if param.param == "Conductivity":
                try:
                    if create_files:
                        return_cnv.to_cnv(
                            f"celltm_{return_cnv.metadata_source.file_name}.cnv"
                        )
                    assert_equal(
                        return_cnv.parameters[param.name].data,
                        pre_cnv.parameters[param.name].data,
                    )
                except AssertionError:
                    assert True
                else:
                    assert False

    def test_loop_edit(self, cnv, create_files):
        instance = LoopRemoval()
        # catch binned data
        try:
            pre_flag = copy.deepcopy(cnv.parameters["flag"].data)
        except KeyError:
            pytest.skip(f"File {cnv.file_name} is binned.")
        return_cnv = instance(
            input=cnv,
            arguments={},
        )
        assert isinstance(return_cnv, CTDData)
        if instance.ran_processing:
            try:
                if create_files:
                    return_cnv.to_cnv(
                        f"loopedit_{return_cnv.metadata_source.file_name}.cnv"
                    )
                assert_equal(pre_flag, return_cnv.parameters["flag"].data)
            except AssertionError:
                assert True
            else:
                assert False

    def test_bin_avg(self, cnv, create_files):
        instance = BinAvg()
        pre_cnv = copy.deepcopy(cnv)
        bin_variable = "prDM"
        if bin_variable not in cnv.parameters:
            pytest.skip()
        try:
            return_cnv = instance(
                input=cnv,
                arguments={
                    "bin_variable": bin_variable,
                    "bin_size": 1,
                },
            )
        except (ValueError, BinnedDataError):
            pytest.skip(f"File {cnv.file_name} is already binned.")
        else:
            if create_files:
                return_cnv.to_cnv(
                    f"bin_avg_{return_cnv.metadata_source.file_name}.cnv",
                    reduced_header=True,
                )
            assert isinstance(return_cnv, CTDData)
            assert (
                return_cnv.parameters.get_data_length()
                < pre_cnv.parameters.get_data_length() / 10
            )



import gsw
from conftest import cnv_path, test_cnv

from ctdam.parser.cnvfile import CnvFile
from ctdam.proc.modules.available_modules import processing_functions
from ctdam.proc.modules.external_functions import (
    ExternalFunctionCaller,
    ExternalFunctions,
)
from ctdam.proc.procedure import Procedure


def test_external_function_dict():
    exfun = ExternalFunctions([gsw])
    assert len(exfun) == 1
    assert len(exfun.get_all_functions()) == 180
    assert exfun.get_all_functions()["rho"]


def test_ex_fun_info():
    info = ExternalFunctions([gsw]).get_all_functions()["Helmholtz_energy_ice"]
    assert info.return_info[0]["desc"]
    cnv = CnvFile(test_cnv)
    success = info.run(cnv)
    assert success
    assert "gsw_Helmholtz_1" in list(cnv.parameters.keys())
    assert (
        cnv.parameters["gsw_Helmholtz_0"].data.shape
        == cnv.parameters["prDM"].data.shape
    )


def test_multi_step_ex_functions():
    gsws = ExternalFunctions([gsw]).get_all_functions()
    step1 = gsws["SA_from_SP_Baltic"]
    step2 = gsws["enthalpy_t_exact"]
    cnv = CnvFile(cnv_path.joinpath("MSM140_1.cnv"))
    success1 = step1.run(cnv)
    assert success1
    success2 = step2.run(cnv)
    assert success2
    assert "gsw_enthalpy_1" in list(cnv.parameters.keys())


def test_caller():
    cnv = CnvFile(cnv_path.joinpath("MSM140_1.cnv"))
    module = ExternalFunctionCaller("SA_from_SP_Baltic", processing_functions)
    module(
        input=cnv,
        output_name=str("test"),
    )
    assert "gsw_saA1" in list(cnv.parameters.keys())


def test_in_procedure():
    procedure = Procedure(
        configuration={
            "input": test_cnv,
            "output_type": "cnvobject",
            "modules": {
                "wildedit_geomar": {},
                "Helmholtz_energy_ice": {},
            },
        }
    )
    assert "gsw_Helmholtz_0" in procedure.ctd_data.parameters



import pytest
from conftest import cnv_path, test_cnv

from ctdam.parser import CnvFile, CnvProcessingSteps


@pytest.fixture
def proc_steps() -> CnvProcessingSteps:
    path_to_file = cnv_path.joinpath("EMB295_14-1.cnv")
    cnv = CnvFile(path_to_file)
    return CnvProcessingSteps(cnv.processing_info)


def test_output_list_length(proc_steps):
    assert len(proc_steps.modules) == 7


def test_individual_module_dict(proc_steps):
    for step, length in {
        "wildedit": 8,
        "datcnv": 6,
        "wfilter": 4,
    }.items():
        assert len(proc_steps.get_step(step).metadata) == length


def test_addition_of_new_proc_lines():
    cnv = CnvFile(test_cnv)
    len_datcnv_metadata = len(
        cnv.processing_steps.get_step("datcnv").metadata.keys()
    )
    cnv.add_processing_metadata("datcnv", "test", "5")
    assert len(cnv.processing_steps.get_names()) == 1
    assert (
        len(cnv.processing_steps.get_step("datcnv").metadata.keys())
        == len_datcnv_metadata + 1
    )



from __future__ import annotations

import logging
import re
from collections import UserDict
from typing import Tuple

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


class Parameters(UserDict):
    """
    A collection of all the parameters in a CnvFile.

    Allows for a much cleaner handling of parameter data and their metadata.
    Will be heavily expanded.

    Parameters
    ----------
    data: list
        The raw data as extraced by DataFile
    metadata: list
        The raw metadata as extraced by DataFile

    Returns
    -------

    """

    def __init__(
        self,
        data: list,
        metadata: list,
        only_header: bool = False,
        bad_flag: float = -9.990e-29,
    ):
        self.data = {}
        self.sample_rate = self.get_sample_rate()
        self.bad_flag = bad_flag
        parsed_metadata, self.duplicate_columns = self.reading_data_header(
            metadata
        )
        if not only_header:
            self.full_data_array = self.create_full_ndarray(data)
            self.create_parameter_instances(
                self.full_data_array, parsed_metadata
            )

    def __repr__(self) -> str:
        return ", ".join(self.data.keys())

    def __eq__(self, other: object, /) -> bool:
        if hasattr(other, "data"):
            try:
                return self.data == other.data
            except ValueError:
                return False
        else:
            return False

    def get_param_types(self) -> list[str]:
        return [p.param for p in self.data.values()]

    def get_data_length(self) -> int:
        return self.data[list(self.data.keys())[0]].data.shape[0]

    def get_full_data_array(self) -> np.ndarray:
        return np.array([parameter.data for parameter in self.data.values()]).T

    def get_names(self) -> list[str]:
        return [parameter.name for parameter in self.data.values()]

    def get_metadata(self) -> dict[str, dict]:
        return {
            parameter.name: parameter.metadata
            for parameter in self.data.values()
        }

    def get_parameter_list(self) -> list[Parameter]:
        """ """
        return list(self.data.values())

    def set_sample_rate(self, rate: float, unit: str):
        self.binned = True
        self.sample_rate = rate
        self.bin_unit = unit

    def get_sample_rate(self, raw_interval_info: str = "") -> float:
        self.binned = False
        self.bin_unit = "seconds"
        try:
            interval_info = raw_interval_info.split(":")
        except Exception:
            pass
        else:
            if interval_info[0] == "seconds":
                sample_rate = np.round(1 / float(interval_info[1]))
                if sample_rate == 1:
                    self.binned = True
            elif interval_info[0] == "decibars":
                sample_rate = float(interval_info[1])
                self.binned = True
                self.bin_unit = "decibars"
            else:
                logger.debug(f"Unknown sample rate {interval_info[0]}.")

        # default rate, calculated if able
        try:
            return sample_rate
        except Exception:
            try:
                sample_rate = float(
                    np.round(1 / np.mean(np.diff(self.data["timeS"])))
                )
            except Exception:
                sample_rate = 24
            return sample_rate

    def create_full_ndarray(self, data_table: list = []) -> np.ndarray:
        """
        Builds a numpy array representing the data table in a cnv file.

        Parameters
        ----------
        data_table: list :
            The data to work with
             (Default value = [])

        Returns
        -------
        A numpy array of the same shape as the cnv files data table

        """
        n = 11
        row_list = []
        for line in data_table:
            row_list.append(
                [
                    line[i : i + n].split()[0]
                    for i in range(0, len(line) - n, n)
                ]
            )
        try:
            return np.array(row_list, dtype=float)
        except ValueError as error:
            logger.error(
                f"Parameters: Wrong data or file format for numpy array creation: {error}"
            )
            return np.array([])

    def sort_parameters(
        self,
        top: list = [
            "depSM",
            "prDM",
            "t090C",
            "t190C",
            "sal00",
            "sal11",
            "sbox0Mm/Kg",
            "sbox1Mm/Kg",
            "flECO-AFL",
            "turbWETntu0",
            "par",
            "spar",
        ],
        bottom: list = [
            "gsw_densityA0",
            "gsw_densityA1",
            "gsw_saA0",
            "gsw_saA1",
            "gsw_ctA0",
            "gsw_ctA1",
            "sbeox0ML/L",
            "sbeox1ML/L",
            "c0mS/cm",
            "c1mS/cm",
            "latitude",
            "longitude",
            "flag",
        ],
    ) -> dict:
        # ensure parameters at the top
        new_data = {}
        for shortname in top:
            for param in self.data.values():
                if shortname == param.name:
                    new_data[shortname] = param

        # ensure parameters at the bottom
        bottom_data = {}
        for shortname in bottom:
            for param in self.data.values():
                if shortname == param.name:
                    bottom_data[shortname] = param

        for param in self.data.values():
            if param.name not in [*top, *bottom]:
                new_data[param.name] = param

        self.data = {**new_data, **bottom_data}
        return self.data

    def create_parameter_instances(
        self,
        array_data: np.ndarray,
        metadata: dict[str, dict],
    ) -> dict[str, Parameter]:
        """
        Differentiates the individual parameter columns into separate parameter
        instances.

        Parameters
        ----------
        metadata: dict[str, dict] :
            The structured metadata dictionary
             (Default value = {})

        Returns
        -------
        A dictionary of parameter instances

        """
        parameter_dict = {}
        list_of_metadata_shortnames = list(metadata.keys())
        # if column number and metadata number is different, we are propably
        # working with duplicate_columns and will drop the duplicates
        if array_data.shape[1] != len(list_of_metadata_shortnames):
            array_data = np.delete(array_data, self.duplicate_columns, 1)
            assert array_data.shape[1] == len(list_of_metadata_shortnames)
        for i in range(array_data.shape[1]):
            key = list_of_metadata_shortnames[i]
            parameter_dict[key] = self.create_parameter(
                data=array_data[:, i],
                metadata=metadata[key],
                name=key,
            )
        return parameter_dict

    def _form_data_table_info(self, output_spans: bool = True) -> list:
        """Recreates the data table descriptions, like column names and spans
        from the structured dictionaries these values were stored in."""
        new_table_info = []
        # 'data table stats'
        data_array = self.get_full_data_array()
        new_table_info.append(f"nquan = {data_array.shape[1]}\r\n")
        new_table_info.append(f"nvalues = {data_array.shape[0]}\r\n")
        new_table_info.append("units = specified\r\n")
        # 'data tables names'
        for index, metadata in enumerate(self.get_metadata().values()):
            new_table_info.append(
                f"name {index} = {metadata['shortname']}: {metadata['longinfo']}\r\n"
            )
        # 'data table spans'
        if output_spans:
            output_formats = [p.output_format for p in self.values()]
            for index, ((minimum, maximum), output_format) in enumerate(
                zip(self.get_spans(), output_formats)
            ):
                try:
                    new_table_info.append(
                        f"span {index} = {output_format.format(minimum)}, {output_format.format(maximum)}\r\n"
                    )
                except ValueError:
                    new_table_info.append(f"span {index} = None, None")
        return new_table_info

    def add_parameter(self, parameter: Parameter, position: str = ""):
        """
        Adds one parameter instance to the collection.

        Parameters
        ----------
        parameter: Parameter :
            The new parameter

        """
        # add to parameter dict at given position
        if position:
            new_dict = {}
            for key, value in self.data.items():
                new_dict[key] = value
                if key == position:
                    new_dict[parameter.name] = parameter
            self.data = new_dict

        else:
            self.data[parameter.name] = parameter

    def create_parameter(
        self,
        data: np.ndarray | int | float | str | None,
        metadata: dict = {},
        name: str = "",
        position: str = "",
    ) -> Parameter:
        """
        Creates a new parameter instance with the given data and metadata.

        The input data is either a numpy array or a single value. The single
        value will be broadcasted to the shape of the data table. A use-case
        would be the addition of an 'event' or 'cast' column.

        Parameters
        ----------
        data: np.ndarray | int | float | str :
            Data to use or expand

        metadata: dict :
            Metadata for the new parameter
             (Default value = {})
        name: str :
            Name to use for missing metadata values
             (Default value = "")

        Returns
        -------
        The new parameter instance

        """
        if len(metadata) < 5:
            if len(name) > 0:
                metadata = self.add_default_metadata(
                    name=name, metadata=metadata
                )
            else:
                raise ValueError(
                    "Please specify either a name or sufficient metadata"
                )
        if not isinstance(data, np.ndarray):
            data = np.full(
                fill_value=data,
                shape=self.get_full_data_array().shape[0],
            )
        parameter = Parameter(
            data=data, metadata=metadata, bad_flag=self.bad_flag
        )
        self.add_parameter(parameter, position)
        return parameter

    def add_default_metadata(
        self,
        name: str,
        metadata: dict = {},
        list_of_keys: list = [
            "shortname",
            "longinfo",
            "name",
            "metainfo",
            "unit",
        ],
    ) -> dict:
        """
        Fills up missing metadata points with a default value.

        Parameters
        ----------
        name: str :
            The value to use as default
        metadata: dict :
            The present metadata
             (Default value = {})
        list_of_keys: list :
             The expected metadata keys

        Returns
        -------
        The full metadata dictionary

        """
        default = {}
        for key in list_of_keys:
            if key not in list(metadata.keys()):
                if key in ["metainfo", "unit"]:
                    default[key] = ""
                default[key] = name
        return {**metadata, **default}

    def update_spans(self):
        """Updates all spans of the parameters."""
        for parameter in self.get_parameter_list():
            parameter.update_span()

    def get_spans(self) -> list[tuple[int, int]]:
        """Returns all span tuples of the parameters."""
        # update spans first
        self.update_spans()
        return [parameter.span for parameter in self.get_parameter_list()]

    def get_pandas_dataframe(self) -> pd.DataFrame:
        """Returns a pandas DataFrame of the current parameter data."""
        data = np.array(
            [parameter.data for parameter in self.get_parameter_list()]
        ).T
        columns = [parameter.name for parameter in self.get_parameter_list()]
        assert data.shape[1] == len(columns)
        df = pd.DataFrame(data=data, columns=columns)
        for column in df.columns:
            try:
                df[column].astype("float64")
            except (TypeError, ValueError):
                df[columns].astype("str")
        return df

    def with_name_type(self, name_type: str = "shortname"):
        """
        Uses the given name_type as column descriptors.

        Parameters
        ----------
        name_type: str :
            The metadata name to use
             (Default value = "shortname")

        """
        for parameter in self.get_parameter_list():
            parameter.use_name(name_type)

    def reading_data_header(
        self, header_info: list = []
    ) -> Tuple[dict[str, dict], list[int]]:
        """Reads the tables header data from the header.

        Parameters
        ----------
        header_info : list:
            the header values from the file
        header_info: list :
             (Default value = [])

        Returns
        -------


        """
        table_header = {}
        duplicate_columns = []
        for line in header_info:
            if line.startswith("name"):
                header_meta_info = {}
                # get basic shortname and the full, non-differentiated info
                shortname = longinfo = line_info = line.split("=", 1)[
                    1
                ].strip()
                try:
                    shortname, longinfo = line_info.split(":", 1)
                except IndexError:
                    pass
                finally:
                    shortname = shortname.strip()
                    if shortname in list(table_header.keys()):
                        try:
                            duplicate_columns.append(
                                int(line.split("=", 1)[0].strip().split()[1])
                            )
                        except IndexError as error:
                            logger.error(
                                f"Could not resolve duplicate column: {
                                    shortname
                                }, {error}"
                            )
                    else:
                        header_meta_info["shortname"] = shortname
                        header_meta_info["longinfo"] = longinfo.strip()
                        metainfo = self._extract_data_header_meta_info(
                            longinfo.strip()
                        )
                        header_meta_info = {**header_meta_info, **metainfo}
                        table_header[shortname.strip()] = header_meta_info
            elif line.startswith("interval"):
                self.sample_rate = self.get_sample_rate(
                    line.split("=", 1)[1].strip()
                )

        return table_header, duplicate_columns

    def _extract_data_header_meta_info(self, line: str) -> dict:
        """Extracts the individual information bits inside of the header lines

        Parameters
        ----------
        line : str:
            one header line, trimmed by the 'name =' prefix and the shortname
        line: str :


        Returns
        -------


        """
        regex_string = r"(?:(?P<name0>.+),\s(?P<metainfo0>.+)\s\[(?P<unit0>.+)\]|(?P<name2>.+)\s\[(?P<unit2>.+)\]|(?P<name3>.+),\s(?P<metainfo2>.[^\s]+)|(?P<name4>.+))"
        regex_check = re.search(regex_string, line, flags=re.IGNORECASE)
        if regex_check:
            regex_info = dict(regex_check.groupdict())
            regex_info = {
                key[:-1]: value
                for key, value in regex_info.items()
                if value is not None
            }
            if len(regex_info) > 2:
                # check for second sensors and adjust their names
                if regex_info["metainfo"][-1] == "2":
                    regex_info["name"] = regex_info["name"] + " 2"
                    regex_info["metainfo"] = regex_info["metainfo"][:-1]
                    if len(regex_info["metainfo"]) == 0:
                        regex_info.pop("metainfo")
            if regex_info["name"] == "flag":
                regex_info["metainfo"] = regex_info["name"]
                regex_info["unit"] = regex_info["name"]
            return regex_info
        return {}


class Parameter:
    """A representation of one parameter in a cnv file.

    Consists of the values of the parameter as well as the metadata.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(
        self,
        data: np.ndarray,
        metadata: dict,
        bad_flag: float = -9.990e-29,
    ) -> None:
        self.data = data
        self.metadata = metadata
        self.name = metadata["shortname"]
        self.param = re.split(r"[,\s]", metadata["name"])[0]
        self.sensor_number = 2 if metadata["name"][-1] == "2" else 1
        self.unit = metadata["unit"]
        self.type = "data" if self.data.dtype in ["float", "int"] else "meta"
        self.bad_flag = bad_flag
        self.parse_to_float()
        self.update_span()
        self.set_output_format()

    def __str__(self) -> str:
        return str(self.metadata["longinfo"])

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other) -> bool:
        return self.data == other.data

    def __len__(self) -> int:
        return self.data.shape[0]

    def get_pandas_series(self) -> pd.Series:
        """Returns a pandas Series of the current parameter data."""
        return pd.Series(data=self.data, name=self.name)

    def use_name(self, name_type: str = "shortname"):
        """
        Uses the given name as parameter descriptor.

        Parameters
        ----------
        name_type: str :
            The metadata name to use
             (Default value = "shortname")

        """
        try:
            self.name = self.metadata[name_type]
        except KeyError:
            return

    def parse_to_float(self):
        """
        Tries to parse the data array type to float.
        """
        try:
            self.data = self.data.astype("float64")
        except ValueError:
            pass

    def update_span(self):
        """
        Updates the data span.

        Uses the first value if dtype is not numeric.
        """
        if self.data.dtype in ["float64", "int"]:
            try:
                mx = np.ma.masked_array(
                    self.data, mask=self.data == self.bad_flag
                )
                self.span = (np.nanmin(mx), np.nanmax(mx))
            except ValueError:
                self.span = (0, 0)
        else:
            self.span = (self.data[0], self.data[0])

    def set_output_format(self):
        if self.name in ["timeU", "flag"]:
            decimal_digits = 0
        elif self.name in ["timeS", "prDM", "sbox0Mm/Kg", "sbox1Mm/Kg"]:
            decimal_digits = 3
        elif self.name in ["latitude", "longitude"]:
            decimal_digits = 5
        elif self.param == "Conductivity":
            decimal_digits = 6
        else:
            decimal_digits = 4
        self.output_format = f"{{:.{decimal_digits}f}}"



import re
from pathlib import Path

from ctdam.parser import DataFile, XMLCONFile


class HexFile(DataFile):
    """
    A representation of a .hex file as used by SeaBird.

    Parameters
    ----------
    path_to_file: Path | str:
        the path to the file
    """

    def __init__(
        self,
        path_to_file: Path | str,
        path_to_xmlcon: Path | str = "",
        *args,
        **kwargs,
    ):
        # force loading only metadata
        super().__init__(path_to_file, True)
        self.xmlcon = self.get_corresponding_xmlcon(path_to_xmlcon)

    def get_corresponding_xmlcon(
        self,
        path_to_xmlcon: Path | str = "",
    ) -> XMLCONFile | None:
        """
        Finds the best matching .xmlcon file inside the same directory.

        Parameters
        ----------
        path_to_xmlcon: Path | str:
            A fixed path to a xmlcon file. Will be checked.
        """
        # xmlcon path given, test and use it
        if isinstance(path_to_xmlcon, str):
            if path_to_xmlcon:
                return XMLCONFile(path_to_xmlcon)
        else:
            if path_to_xmlcon.exists():
                return XMLCONFile(path_to_xmlcon)
        # no xmlcon path, lets find one in the same dir
        # get all xmlcons in the dir
        # first, try the very same name
        same_name_xmlcon = self.path_to_file.with_suffix(".XMLCON")
        if same_name_xmlcon.exists():
            return XMLCONFile(same_name_xmlcon)

        # otherwise, take the xmlcon of the previous hex file, sorted by name
        # use either the extracted cruise name or the first five letters for
        # searching for xmlcons of the same cruise
        if self.cruise:
            prefix = (
                re.split(r"[_\-\/]", self.cruise.lower())[0]
                if any(x in self.cruise.lower() for x in ["_", "-", "/"])
                else self.cruise.lower()
            )
        else:
            prefix = self.file_name.lower()[:5]
        xmlcons = [
            xmlcon
            for xmlcon in sorted(
                self.file_dir.glob("*.XMLCON", case_sensitive=False)
            )
            if xmlcon.stem.lower().startswith(prefix)
        ]
        if not xmlcons:
            return None

        # looking back and using xmlcon from the previous hex file
        all_hexes = [
            hex
            for hex in sorted(
                self.file_dir.glob("*.hex", case_sensitive=False)
            )
            if hex.stem.lower().startswith(prefix)
        ]
        index = all_hexes.index(self.path_to_file)
        previous_hexes = all_hexes[:index]
        if previous_hexes:
            return HexFile(previous_hexes[-1]).xmlcon

        return XMLCONFile(xmlcons[0])



import pytest
from conftest import (
    check_and_remove_file,
    hex_path,
)

from ctdam.parser.ctddata import CTDData

try:
    from ctdam.entry.cli import batch, convert
except ImportError:
    pass
else:

    @pytest.mark.long
    def test_convert_function(tmp_path):
        pattern = "SO308-2"
        result = convert(
            input_dir=str(hex_path),
            output_dir=str(tmp_path),
            pattern=pattern,
        )
        assert len(result) == 3
        for ctd_data in result:
            check_and_remove_file(
                tmp_path.joinpath(ctd_data.file_name).with_suffix(".cnv")
            )

    @pytest.mark.long
    def test_batch_processing_function():
        pattern = "SO308-2*.hex"
        proc_config = {
            "output_type": "internal",
            "modules": {
                "wildedit_geomar": {},
                "celltm": {},
            },
        }
        result = batch(
            input_dir=str(hex_path),
            config=proc_config,
            pattern=pattern,
        )
        assert len(result) == 3
        for file in result:
            assert isinstance(file, CTDData)
            assert len(file.processing_steps) == 3



import logging
import sys
import tomllib
from math import floor
from pathlib import Path

import gsw
import numpy as np
import pandas as pd
from seabirdscientific import cal_coefficients as sbs_cal
from seabirdscientific import conversion as sbs_con

from ctdam.parser.parameter import Parameters

logger = logging.getLogger(__name__)


class ParameterMapping:
    def __init__(
        self,
        xmlcon_part: dict,
        raw_data: pd.DataFrame,
        parameters: Parameters,
        voltage_sensors_start: int = 6,
    ) -> None:
        self.source = xmlcon_part
        self.name = (
            xmlcon_part["SensorName"]
            if xmlcon_part["SensorName"]
            else xmlcon_part["SerialNumber"]
        )
        self.sensor_id = xmlcon_part["@SensorID"]
        self.raw_data = raw_data
        self.voltage_sensors_start = voltage_sensors_start
        sensor_mapping_file = Path(__file__).parent.joinpath(
            "sensor_mapping.toml"
        )
        if not sensor_mapping_file.exists():
            sys.exit(
                f"No sensor mapping file found. Looked in {sensor_mapping_file}. Aborting."
            )
        with open(sensor_mapping_file, "rb") as file:
            self.mapper = tomllib.load(file)
        if self.name[-1] == "2":
            self.name = self.name[:-1]
            self.second_sensor = True
        else:
            self.second_sensor = False
        self.parameters = parameters
        self.param_types = [p.param for p in parameters.get_parameter_list()]
        self.sample_interval = 1 / 24
        self.sensor_data = self.locate_sensor_data(raw_data)
        if isinstance(self.sensor_data, np.ndarray):
            self.coefficients = self.extract_coefficients(xmlcon_part)
            self.metadata = self.map_metadata()

    def extract_coefficients(self, source: dict):
        # Temperature SBE 4
        if self.sensor_id == "55":
            if source["UseG_J"] == "1":
                self.convert_freq_temperature()
            else:
                self.convert_temperature()

        # Conductivity SBE 3
        elif self.sensor_id == "3":
            if source["Coefficients"][0]["@equation"] == "1":
                logger.error(
                    "Unsupported Conductivity coefficients given, please use the new coefficients."
                )
            self.convert_conductivity()
        # Pressure
        elif self.sensor_id == "45":
            self.convert_pressure()
        # Oxygen SBE43
        elif self.sensor_id == "38":
            if source["CalibrationCoefficients"][0]["@equation"] == "1":
                logger.error(
                    "Unsupported Oxygen coefficients given, please use the new coefficients."
                )
            self.convert_oxygen()
        # PAR
        elif self.sensor_id == "42":
            self.coef = sbs_cal.PARCoefficients
            self.coef.im = float(source["M"])
            self.coef.a0 = float(source["CalibrationConstant"])
            self.coef.a1 = float(source["Offset"])
            self.coef.multiplier = float(source["Multiplier"])
            self.converted_data = (
                self.coef.multiplier
                * (
                    (10**9 * 10 ** (self.sensor_data / self.coef.im))
                    / self.coef.a0
                )
                + self.coef.a1
            )

        # Altimeter
        elif self.sensor_id == "0":
            self.coef = sbs_cal.AltimeterCoefficients
            self.coef.slope = float(source["ScaleFactor"])
            self.coef.offset = float(source["Offset"])
            self.converted_data = sbs_con.convert_altimeter(
                volts=self.sensor_data,
                coefs=self.coef,
            )
        # Fluorometer
        elif self.sensor_id in ["20", "19"]:
            self.coef = sbs_cal.ECOCoefficients
            self.coef.slope = float(source["ScaleFactor"])
            self.coef.offset = float(source["Vblank"])
            self.converted_data = sbs_con.convert_eco(
                raw=self.sensor_data,
                coefs=self.coef,
            )
        # Turbidity Meter
        elif self.sensor_id == "67":
            self.coef = sbs_cal.ECOCoefficients
            self.coef.slope = float(source["ScaleFactor"])
            self.coef.offset = float(source["DarkVoltage"])
            self.converted_data = sbs_con.convert_eco(
                raw=self.sensor_data,
                coefs=self.coef,
            )
        # SPAR
        elif self.sensor_id == "51":
            self.coef = sbs_cal.SPARCoefficients
            self.coef.conversion_factor = float(source["ConversionFactor"])
            self.converted_data = sbs_con.convert_spar_biospherical(
                volts=self.sensor_data,
                coefs=self.coef,
            )
        # User Polynomial Sensor
        elif self.sensor_id == "61":
            if (
                self.name.startswith("Oxygen")
                or self.source["SerialNumber"] == "Pyro1"
            ):
                self.convert_oxygen_pyro_science()
            elif self.name.startswith("Flow Meter"):
                self.create_parameter(
                    self.sensor_data * float(self.source["A1"]),
                    self.map_metadata("Flow Meter"),
                    "Flow",
                )

    def locate_sensor_data(self, raw_data: dict) -> np.ndarray | None:
        if self.name in ["Temperature", "Conductivity"]:
            name = self.name.lower()
            if self.second_sensor:
                name = "secondary " + name

        elif self.name == "Pressure":
            name = "digiquartz pressure"
        elif self.name == "SPAR":
            name = "surface par"
        else:
            sensor_index = int(self.source["Channel"])
            name = f"volt {sensor_index - self.voltage_sensors_start}"

        try:
            sensor_data = raw_data[name].values
        except Exception as error:
            logger.debug(
                f"Could not locate sensor data for {self.name}: {error}"
            )
            sensor_data = None
        return sensor_data

    def create_parameter(self, data, metadata: dict = {}, name: str = ""):
        try:
            self.parameters.create_parameter(
                data=data,
                metadata=metadata,
                name=name,
            )
        except AttributeError:
            logger.debug(f"{name} had no succesfull mapping.")

    def convert_freq_temperature(self):
        self.coef = sbs_cal.TemperatureFrequencyCoefficients
        for param in ["G", "H", "I", "J", "F0"]:
            setattr(self.coef, param.lower(), float(self.source[param]))
        self.converted_data = sbs_con.convert_temperature_frequency(
            frequency=self.sensor_data,
            coefs=self.coef,
            standard="ITS90",
            units="C",
        )

    def convert_temperature(self):
        self.coef = sbs_cal.TemperatureCoefficients
        for index, param in enumerate(["A", "B", "C", "D"]):
            setattr(self.coef, f"a{index}", float(self.source[param]))
        self.converted_data = sbs_con.convert_temperature(
            frequency=self.sensor_data,
            coefs=self.coef,
            standard="ITS90",
            units="C",
        )

    def convert_pressure(self):
        self.coef = sbs_cal.PressureDigiquartzCoefficients
        for param in self.source:
            if param in [
                "Channel",
                "SensorName",
                "@SensorID",
                "SerialNumber",
                "CalibrationDate",
                "Slope",
                "Offset",
            ]:
                continue
            setattr(
                self.coef,
                param if param.startswith("A") else param.lower(),
                float(self.source[param]),
            )
        self.converted_data = sbs_con.convert_pressure_digiquartz(
            pressure_count=self.sensor_data,
            compensation_voltage=self.raw_data["temperature compensation"]
            .astype(float)
            .values,
            coefs=self.coef,
            units="dbar",
            sample_interval=self.sample_interval,
        )
        # apply correction on data
        self.converted_data = self.converted_data * float(
            self.source["Slope"]
        ) + float(self.source["Offset"])

    def convert_conductivity(self):
        self.coef = sbs_cal.ConductivityCoefficients
        for param in ["G", "H", "I", "J", "CPcor", "CTcor", "WBOTC"]:
            setattr(
                self.coef,
                param.lower(),
                float(self.source["Coefficients"][1][param]),
            )
        if "Pressure" not in self.param_types:
            return
        p_values = self.parameters["prDM"].data
        if "Temperature" in self.param_types:
            if self.second_sensor:
                t_values = self.parameters["t190C"].data
            else:
                t_values = self.parameters["t090C"].data
        else:
            return

        self.converted_data = sbs_con.convert_conductivity(
            conductivity_count=self.sensor_data,
            temperature=t_values,
            pressure=p_values,
            coefs=self.coef,
            scalar=1,
        )
        self.convert_salinity(self.converted_data, t_values, p_values)

    def convert_salinity(
        self,
        conductivity: np.ndarray,
        t_values: np.ndarray,
        p_values: np.ndarray,
    ):
        # TODO: allow selection of baltic salinity conversion here
        converted_data = gsw.SP_from_C(
            C=conductivity,
            t=t_values,
            p=p_values,
        )
        self.create_parameter(converted_data, self.map_metadata("Salinity"))

    def convert_sbe43_oxygen(
        self,
        voltage: np.ndarray,
        temperature: np.ndarray,
        pressure: np.ndarray,
        salinity: np.ndarray,
        coefs: sbs_cal.Oxygen43Coefficients,
        apply_tau_correction: bool = False,
        apply_hysteresis_correction: bool = False,
        window_size: float = 1,
        sample_interval: float = 1,
    ):
        """Overwrite of Sea-Birds super slow function."""
        # start with all 0 for the dvdt
        dvdt_values = np.zeros(len(voltage))
        if apply_tau_correction:
            # Calculates how many scans to have on either side of our median
            # point, accounting for going out of index bounds
            scans_per_side = floor(window_size / 2 / sample_interval)
            for i in range(scans_per_side, len(voltage) - scans_per_side):
                ox_subset = voltage[
                    i - scans_per_side : i + scans_per_side + 1
                ]

                time_subset = np.arange(
                    0,
                    len(ox_subset) * sample_interval,
                    sample_interval,
                    dtype=float,
                )

                def manual_linregress(x, y):
                    x_mean, y_mean = np.mean(x), np.mean(y)
                    cov = np.sum((x - x_mean) * (y - y_mean))
                    var = np.sum((x - x_mean) ** 2)
                    slope = cov / var
                    intercept = y_mean - slope * x_mean
                    return slope, intercept

                slope, _ = manual_linregress(time_subset, ox_subset)

                dvdt_values[i] = slope

        correct_ox_voltages = voltage.copy()
        if apply_hysteresis_correction:
            # Hysteresis starts at 1 because 0 can't be corrected
            for i in range(1, len(correct_ox_voltages)):
                # All Equation info from APPLICATION NOTE NO. 64-3
                d = 1 + coefs.h1 * (np.exp(pressure[i] / coefs.h2) - 1)
                c = np.exp(-1 * sample_interval / coefs.h3)
                ox_volts = correct_ox_voltages[i] + coefs.v_offset

                prev_ox_volts_new = correct_ox_voltages[i - 1] + coefs.v_offset
                ox_volts_new = (
                    (ox_volts + prev_ox_volts_new * c * d)
                    - (prev_ox_volts_new * c)
                ) / d
                ox_volts_final = ox_volts_new - coefs.v_offset
                correct_ox_voltages[i] = ox_volts_final

        oxygen = sbs_con._convert_sbe43_oxygen(
            correct_ox_voltages,
            temperature,
            pressure,
            salinity,
            coefs,
            dvdt_values,
        )
        return oxygen

    def convert_oxygen(self):
        self.coef = sbs_cal.Oxygen43Coefficients
        for param, value in self.source["CalibrationCoefficients"][1].items():
            param = f"v_{param}" if param == "offset" else param
            param = "tau_20" if param == "Tau20" else param
            setattr(self.coef, param.lower(), float(value))
        if "Pressure" not in self.param_types:
            return
        p_values = self.parameters["prDM"].data
        if "Temperature" in self.param_types:
            if self.second_sensor:
                t_values = self.parameters["t190C"].data
            else:
                t_values = self.parameters["t090C"].data
        else:
            return
        if "Salinity" in self.param_types:
            if self.second_sensor:
                s_values = self.parameters["sal11"].data
            else:
                s_values = self.parameters["sal00"].data
        else:
            return
        converted_data = self.convert_sbe43_oxygen(
            voltage=self.sensor_data,
            temperature=t_values,
            pressure=p_values,
            salinity=s_values,
            coefs=self.coef,
            apply_tau_correction=True,
            apply_hysteresis_correction=True,
            window_size=2,
            sample_interval=self.sample_interval,
        )
        self.create_parameter(
            converted_data,
            self.map_metadata("Oxygen mlL"),
        )
        # TODO: flexibilize this
        # give out umol/kg
        absolute_salinity = gsw.SA_from_SP(
            SP=s_values,
            p=p_values,
            lon=self.raw_data["NMEA Longitude"],
            lat=self.raw_data["NMEA Latitude"],
        )
        self.create_parameter(
            absolute_salinity,
            self.map_metadata("Absolute Salinity"),
        )
        conservative_temperature = gsw.conversions.CT_from_t(
            SA=absolute_salinity,
            t=t_values,
            p=p_values,
        )
        self.create_parameter(
            conservative_temperature,
            self.map_metadata("Conservative Temperature"),
        )
        # TODO: flexibile sigma selection
        potential_density = gsw.density.sigma0(
            SA=absolute_salinity,
            CT=conservative_temperature,
        )
        self.create_parameter(potential_density, self.map_metadata("Density"))
        self.converted_data = sbs_con.convert_oxygen_to_umol_per_kg(
            ox_values=converted_data,
            potential_density=potential_density,
        )

    def convert_oxygen_pyro_science(self):
        try:
            oxygen = self.sensor_data * float(self.source["A1"])
        except TypeError as error:
            logger.error(
                f"could not convert pyro oxygen: {self.source['SerialNumber']}: {error}"
            )
        else:
            # oxygen in umol/kg
            self.create_parameter(oxygen, self.map_metadata("Oxygen Pyro 2"))
            # oxygen in ml/l
            if "ConservativeTemperature" in self.param_types:
                conservative_temperature = self.parameters["gsw_ctA0"].data
            else:
                return
            if "AbsoluteSalinity" in self.param_types:
                absolute_salinity = self.parameters["gsw_saA0"].data
            else:
                return
            potential_density = gsw.density.sigma0(
                SA=absolute_salinity,
                CT=conservative_temperature,
            )
            oxygen_mll = (
                (oxygen * (potential_density + 1000))
                / 1
                / sbs_con.OXYGEN_MLPERL_TO_UMOLPERKG
            )
            self.create_parameter(
                oxygen_mll,
                self.map_metadata("Oxygen Pyro mlL 2"),
            )

    def map_metadata(self, name: str = "") -> dict:
        name = name if name else self.name
        if self.second_sensor:
            name = name + " 2"
        try:
            return self.mapper["metadata"][name]
        except KeyError:
            return {}



import logging
from pathlib import Path
from typing import Tuple

import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import uniform_filter1d

from ctdam.parser import CnvFile
from ctdam.parser.ctddata import CTDData
from ctdam.proc.module import ArrayModule
from ctdam.proc.utils import is_directly_measured_value

logger = logging.getLogger(__name__)


class WildeditGEOMAR(ArrayModule):
    """
    Flags outliers in a dataset via standard deviation.

    Iterates over blocks of data, calculates mean and standard deviation and
    flags data outside a pre-set standard deviation window. In contrast to the
    standard SeaBird processing module, wild_edit, this module uses a sliding
    window around each data point instead of fixed blocks of data.
    Additionally, while the SeaBird variant terminates after two fixed cycles
    of flagging, here we iterate as long as there are bad values found.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "wildedit_geomar"

    def __call__(
        self,
        input: Path | str | CnvFile | CTDData,
        arguments: dict = {},
        output: str = "cnvobject",
        output_name: str | None = None,
        default_values: dict = {
            "std1": 3.0,
            "std2": 10.0,
            "window_size": 50,
            "minstd": 0,
            "flag_points": True,
        },
        **kwargs,
    ) -> None | CnvFile | CTDData:
        return super().__call__(
            input, arguments, output, output_name, default_values
        )

    def transformation(self) -> bool:
        """
        Selects the appropiate data columns and applies the flagging to each
        one individually.

        Returns
        -------
        A numpy array, representing full cnv data with outliers removed.

        """
        for key, value in self.arguments.items():
            try:
                self.arguments[key] = float(value)
            except ValueError:
                self.arguments[key] = value
            if key == "window_size":
                self.arguments[key] = int(value)
        return_value = False
        all_wildedit_flags = np.zeros_like(
            self.ctd_data["flag"].data, dtype=bool
        )
        for param in self.ctd_data:
            if is_directly_measured_value(param):
                new_data, new_flag = wildedit_geomar(
                    data=param.data,
                    flag=self.ctd_data["flag"].data,
                    **self.arguments,
                )
                param.data = new_data
                if not self.arguments["flag_points"]:
                    all_wildedit_flags |= new_flag
                return_value = True
        return return_value


def wildedit_geomar(
    data: np.ndarray,
    flag: np.ndarray = np.ndarray(shape=(0, 0)),
    std1: float = 3.0,
    std2: float = 10.0,
    window_size: int = 50,
    minstd: float = 0,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Wild editing filter, addapted from Gerd Krahmann.

    Uses a sliding window and repeated flagging loops until no bad values are
    found.

    Parameters
    ----------
    data: np.ndarray :
        The input data array.

    flag: np.ndarray :
        The data flag array.

    std1: float :
        The standard deviation cutoff for the first flagging loop.
         (Default value = 3.0)
    std2: float :
        The standard deviation cutoff for the all the following loops.
         (Default value = 10.0)
    window_size: int :
        The size of the sliding window.
         (Default value = 50)
    minstd: float :
        The minimum standard deviation threshold to flag data.
         (Default value = 0)

    Returns
    -------
    A tuple of cleaned data and new data flags.

    """

    if window_size < 10:
        raise ValueError("window_size must be >= 10")

    data = np.asarray(data, dtype=float)

    if flag.size > 0:
        flag = np.asarray(flag)

        good_indices = np.where(flag == 1)[0]
        if good_indices.size > 1:
            interp_func = interp1d(
                good_indices,
                data[good_indices],
                kind="nearest",
                fill_value="extrapolate",
                bounds_error=False,
            )
            data = interp_func(np.arange(len(data)))

    def running_mean_std(
        data: np.ndarray, window_size: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Fast running mean and standard deviation using uniform filter."""
        mean = uniform_filter1d(data, size=window_size, mode="nearest")
        mean_sq = uniform_filter1d(data**2, size=window_size, mode="nearest")
        var = mean_sq - mean**2
        var = np.maximum(0, var)
        std = np.sqrt(var)
        return mean, std

    def flag_bad_points(
        x: np.ndarray,
        mean: np.ndarray,
        std: np.ndarray,
        threshold: float,
        min_std: float,
    ) -> np.ndarray:
        """Apply mean and standard deviation to find outliers."""
        std = np.abs(std)
        if len(std) > 100:
            sorted_std = np.sort(std)
            low_limit = sorted_std[len(std) // 100]
            std = np.maximum(std, low_limit)
        std = np.maximum(std, min_std)
        return np.abs(x - mean) > threshold * std

    total_bad_mask = np.zeros_like(data, dtype=bool)
    local_data = data.copy()

    threshold = std1
    iteration = 0

    while True:
        iteration += 1

        mean, std = running_mean_std(local_data, window_size)
        bad_mask = flag_bad_points(data, mean, std, threshold, minstd)
        new_bad_mask = bad_mask & ~total_bad_mask

        if not np.any(new_bad_mask):
            break

        total_bad_mask |= new_bad_mask
        good_indices = np.where(~total_bad_mask)[0]

        if good_indices.size < 2:
            break

        interp_func = interp1d(
            good_indices,
            local_data[good_indices],
            kind="cubic",
            fill_value="extrapolate",
            bounds_error=False,
        )
        local_data[new_bad_mask] = interp_func(np.where(new_bad_mask)[0])
        # switch to second threshold after first iteration
        threshold = std2

    data_geomar = data.copy()
    data_geomar[total_bad_mask] = np.nan

    flag_geomar = np.zeros_like(data, dtype=bool)
    flag_geomar[total_bad_mask] = True

    return data_geomar, flag_geomar



from __future__ import annotations

import copy
from collections import UserList


class CnvProcessingSteps(UserList):
    """
    A python representation of the individual ctdam.proc steps conducted
    in the process of a cnv file creation. These modules are stored in
    a dictionary structure, together with all the variables/metadata/etc.
    given in the header of a cnv file.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, raw_processing_info: list):
        self.modules = self.extract_individual_modules(raw_processing_info)
        self.data = []
        for module in self.modules:
            self.data.append(
                self.create_step_instance(module, raw_processing_info)
            )

    def _form_processing_info(self) -> list:
        out_list = []
        for module in self.data:
            module = copy.deepcopy(module)
            if "vars" in module.metadata and module.name != "wildedit":
                module.metadata["date"] = (
                    module.metadata["date"]
                    + f" [{module.name.lower()}_vars = {module.metadata.pop('vars')}]"
                )
            if module.name == "binavg":
                collection_string = module.metadata["binavg_surface_bin"][
                    "surface_bin"
                ]
                for k, v in module.metadata["binavg_surface_bin"].items():
                    if k != "surface_bin":
                        collection_string += f", {k} = {v}"
                module.metadata["binavg_surface_bin"] = collection_string
            for key, value in module.metadata.items():
                if module.name == "wfilter" and key == "action":
                    for action_key, action_value in value.items():
                        out_list.append(
                            f"wfilter_action {action_key} = {action_value}\r\n"
                        )
                else:
                    out_list.append(f"{module.name}_{key} = {value}\r\n")
        out_list.append("file_type = ascii\r\n")
        return out_list

    def get_names(self) -> list[str]:
        return [step.name for step in self.data]

    def extract_individual_modules(self, raw_info: list[str]) -> list:
        """ """
        module_list = []
        for line in raw_info:
            module = line.split("_")[0]
            if (module not in module_list) and (
                line.split()[0] != "file_type"
            ):
                module_list.append(module)
        return module_list

    def create_step_instance(
        self,
        module: str,
        raw_info: list[str],
    ) -> ProcessingStep:
        """

        Parameters
        ----------
        module :


        Returns
        -------

        """
        # TODO: probably need to split this into smaller bits
        out_dict = {}
        inner_action_dict = {}
        # extract lines corresponding to the module
        for line in raw_info:
            if module == line.split("_")[0]:
                # removing the module names from the lines
                shifting_index = len(module) + 1
                line_content = line[shifting_index:]
                # handle the case of the validation methods keyword being
                # 'action', which corresponds to an entire dict of values
                if line_content[:6] == "action":
                    inner_action_dict = self._module_dict_feeder(
                        line_content[6:], inner_action_dict
                    )
                else:
                    # handle the cases where after some date value, another value
                    # is printed inside of [] brackets
                    double_value_list = line_content.split("[")
                    if len(double_value_list) > 1:
                        out_dict = self._module_dict_feeder(
                            double_value_list[1][shifting_index:-2], out_dict
                        )
                        line_content = double_value_list[0]
                    if line_content[:11] == "surface_bin":
                        surface_bin_dict = {}
                        for line in line_content.split(","):
                            self._module_dict_feeder(line, surface_bin_dict)
                        out_dict["surface_bin"] = surface_bin_dict
                        continue
                    # usual behavior, for 99% cases:
                    # assigning key and value to the module dict
                    out_dict = self._module_dict_feeder(line_content, out_dict)
        if inner_action_dict:
            out_dict["action"] = inner_action_dict
        return ProcessingStep(module, out_dict)

    def _module_dict_feeder(
        self,
        line: str,
        dictionary: dict,
        split_value: str = "=",
    ):
        """

        Parameters
        ----------
        line: str :

        dictionary: dict :

        split_value: str :
             (Default value = '=')

        Returns
        -------

        """
        # adds the values of a specific header line into a dictionary
        try:
            key, value = line.split(split_value)
        except ValueError:
            pass
        else:
            dictionary[key.strip()] = value.strip()
        return dictionary

    def get_step(self, step: str) -> ProcessingStep | None:
        """

        Parameters
        ----------
        module: str :


        Returns
        -------

        """
        for index, element in enumerate(self.data):
            if str(element) == step:
                return self.data[index]
        return None

    def add_info(
        self,
        module: str,
        key: str,
        value: str,
    ) -> ProcessingStep | None:
        """
        Adds new processing lines to the list of processing module information

        Parameters
        ----------
        module: str :
            the name of the processing module
        key: str :
            the description of the value
        value: str :
            the information

        Returns
        -------
        the altered ProcessingStep

        """
        if module in self.modules:
            step_info = self.get_step(module)
            if step_info:
                step_info.metadata[key] = value
        else:
            step_info = ProcessingStep(name=module, metadata={key: value})
            self.data.append(step_info)
            self.modules.append(module)
        return step_info


class ProcessingStep:
    """
    Class that is meant to represent one individual processing step, that lead
    to the current status of the cnv file. Can be a custom processing step or
    one of the original Sea-Bird ones.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, name: str, metadata: dict):
        self.name = name
        self.metadata = metadata

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other) -> bool:
        return self.metadata == other.metadata



import shutil
import sys
from pathlib import Path

from ctdam.parser.parameter import Parameter
from ctdam.parser.processing_steps import CnvProcessingSteps


def default_seabird_exe_path() -> Path:
    """Creates a platform-dependent default path to the Sea-Bird exes."""
    exe_path = "Program Files (x86)/Sea-Bird/SBEDataProcessing-Win32/"
    if sys.platform.startswith("win"):
        path_prefix = Path("C:/")
    else:
        path_prefix = Path.home().joinpath(".wine/drive_c")
    return path_prefix.joinpath(exe_path)


def is_directly_measured_value(parameter: Parameter) -> bool:
    """
    Returns whether a parameter has been measured via a sensor or is calculated.
    """
    value_list = [
        "Pressure",
        "Conductivity",
        "Temperature",
        "Oxygen",
        "PAR/Irradiance",
        "SPAR",
        "Fluorescence",
        "Turbidity",
    ]
    return parameter.metadata["name"] in value_list


def get_alignment_delay_and_correlation_values(
    processing_info: CnvProcessingSteps,
) -> list:
    """
    Finds the two numerical values in the processing output produced by the
    custom alignment tool. These are extracted separately for each sensor and
    sorted inside of list[tuple] structure.
    """
    output = []
    sensor = 1
    for step in processing_info:
        if step.name == "alignctd":
            for param, value in step.metadata.items():
                if str(sensor - 1) not in param:
                    continue
                # catch files that have used Sea-Bird align or were run with a
                # pre-set value, these are not interesting for this extraction
                try:
                    delay, metainfo = value.split(",")
                except ValueError:
                    continue
                sensor_tuple = (delay.strip()[:-1], metainfo.strip()[-4:])
                output.append(sensor_tuple)
                sensor += 1
            if sensor > 2:
                break

    return output


def fill_file_type_dir(file_type_dir: Path, file: Path, copy: bool = True):
    """
    Copies the target input and output files into individual type
    directories.

    A 'file type directory' is a directory that is meant to collect all
    the file of the same file extension that accumulate over multiple
    processings. For typical Sea-Bird processings you usually end up with
    something like this:

    root-dir
        - hex
        - cnv
        - XMLCON
        - btl
        - bl
        - hdr

    Parameters
    ----------
    file: Path :

    copy: bool :
            (Default value = True)

    Returns
    -------

    """
    file_dir = file_type_dir.joinpath(file.suffix.strip("."))
    if not file_dir.exists():
        file_dir.mkdir(parents=True)
    new_path = file_dir.joinpath(file.name)
    if copy:
        try:
            shutil.copyfile(file, new_path)
        except shutil.SameFileError:
            pass
    else:
        file.rename(new_path)


class BinnedDataError(Exception):
    """A custom error to throw when binned data has been detected."""

    def __init__(self, file_name: str, step_name: str):
        super().__init__(
            f"{step_name} cannot be applied to binned data in {file_name}"
        )



import logging
import platform
import subprocess
from pathlib import Path
from time import sleep

from ctdam.proc.seabird import ProcessingModule

logger = logging.getLogger(__name__)


class ProcessingStep:
    """
    A representation of a single Seabird processing step.

    One instance of this class corresponds to one Seabird processing module and
    its invocation on one single file.

    Parameters
    ----------

    module: ProcessingModule :
        The Sea-Bird module to run.

    input_path: Path | str :
        The path to the target file.

    xmlcon_path: Path | str | None :
        The path to the XMLCON file.

    output_path: Path | str | None :
        The path to write the new file to. Default is the same as the input
        one.

    original_input_path: Path | str | None :
        The path to the target file of the whole processing workflow.

    new_name: str | None :
        Option to set a new name to the output file.

    verbose: bool :
        Allows to run the module in "window" mode, which gives the option to
        check and manipulate the modules configuration. Is suppressed by
        default.

    Returns
    -------
    A runnable instance.

    """

    def __init__(
        self,
        module: ProcessingModule,
        input_path: Path | str,
        xmlcon_path: Path | str | None = None,
        output_path: Path | str | None = None,
        original_input_path: Path | str | None = None,
        new_name: str | None = None,
        verbose: bool = False,
    ):
        self.exe = module.exe
        self.psa = module.psa
        self.name = module.name
        self.input_path = Path(input_path)
        self.input_dir = self.input_path.parent
        self.new_name = self.build_new_name(new_name, module.new_file_suffix)
        self.verbose = verbose
        if output_path:
            self.output_path = Path(output_path)
        else:
            self.output_path = self.input_dir
        if xmlcon_path:
            self.xmlcon = Path(xmlcon_path)
        else:
            self.xmlcon = self.input_dir.joinpath(
                self.input_path.stem + ".XMLCON"
            )
        if original_input_path:
            self.original_input_path = Path(original_input_path)
        else:
            self.original_input_path = self.input_path

    def __str__(self) -> str:
        return self.name

    def build_new_name(
        self,
        new_name: str | None,
        new_file_suffix: str,
    ) -> Path:
        """
        Sets the output file path with all the information given.

        Parameters
        ----------
        new_name: str | None :
            The optional new output name.

        new_file_suffix: str :
            The optional suffix to append to the new output name.

        Returns
        -------
        A string representing the new name.

        """
        if new_name:
            name = Path(new_name).stem
        else:
            name = self.input_path.stem
        return Path(name + new_file_suffix).with_suffix(".cnv")

    def run_string(self) -> list:
        """
        Builds the command line command that is used to run the module.

        Collects the different bits of information and options and handles them
        accordingly. Does also consider module specific specialities and the
        wanted verbosity.
        """
        run_command_list = [str(self.exe)]
        if not self.verbose:
            run_command_list.append("/s")
            run_command_list.append("/m")
        if self.name == "iow_btl_id":
            run_command_list.append(f"-i{self.input_path}")
            run_command_list.append(f"-o{self.output_path}")
        elif self.name == "bottlesum":
            input_ros_path = self.original_input_path.with_suffix(".ros")
            if not input_ros_path.exists():
                input_ros_path = self.input_path.with_suffix(".ros")
                if not input_ros_path.exists():
                    return []
            run_command_list.append(f"/c{self.xmlcon}")
            run_command_list.append(f"/p{self.psa}")
            run_command_list.append(f"/i{input_ros_path}")
            run_command_list.append(f"/o{self.output_path}")
            run_command_list.append(f"/f{self.new_name.with_suffix('.btl')}")
        else:
            run_command_list.append(f"/p{self.psa}")
            run_command_list.append(f"/i{self.input_path}")
            run_command_list.append(f"/o{self.output_path}")
        if self.name not in ["airpressure", "iow_btl_id", "bottlesum"]:
            run_command_list.append(f"/c{self.xmlcon}")
            run_command_list.append(f"/f{self.new_name}")
        if self.name == "airpressure":
            run_command_list.append(f"/f\\{self.new_name}")
        logger.debug(
            f"Command line arguments of processing step {self.name}:\n\t - "
            + "\n\t - ".join(run_command_list)
        )
        return run_command_list

    def run(self, command: list | None = None, timeout: int = 60):
        """
        Handles the creation of the command line command.

        Is mostly needed to correctly running the Sea-Bird processing modules
        on a linux system using the wine emulator.

        Parameters
        ----------
        command: list | None :
             The command to run. If not given, creates it.
        """
        if not command:
            command = self.run_string()
        # To run processing on Linux, two programs need to be installed, wine
        # and unix2dos. Inside wine, the SeabirdProcessing Software Suite needs
        # to be installed and configured. If airpressure or iow_btl_id want to
        # be used, these need to be copied in the directory where the other
        # processing exes reside.
        if platform.system() == "Linux":
            # seabird cannot work with unix-style line-endings, using only LF,
            # so the input file needs to be converted to use windows-style
            # line-endings, featuring CR followed by a LF
            self.run_process(["unix2dos", self.input_path])
            command.insert(0, "wine")
        self.run_process(command, timeout)

    def run_process(self, command: list, timeout: int = 60):
        """
        Creates and monitors the execution of the module.

        This is done in a separate process, to allow leveraging
        multi-threading. Does also set a timeout for killing the process after
        some time.

        Parameters
        ----------
        command: list :
            The command to run, with all the necessary parameters.

        timeout: int :
            The time in seconds to wait for the execution to finish. Kills the
            process otherwise.
        """
        if not command:
            return
        try:
            ps = subprocess.Popen(
                command,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                shell=platform.platform().startswith("Win"),
            )
        except (
            subprocess.CalledProcessError,
            subprocess.TimeoutExpired,
        ) as error:
            logger.error(
                f"Running processing step of the module {
                    self.name
                } and the command '{command}' failed: {error}"
            )
        else:
            run_time = 0
            while ps.poll() is None and run_time < timeout:
                sleep(1)
                run_time += 1
            if run_time == timeout:
                ps.kill()
                raise subprocess.TimeoutExpired(
                    f"Process {command} ran into a timeout", timeout=timeout
                )



import logging

import numpy as np
import pytest
from conftest import btl_path, cnv_path, test_cnv
from numpy.testing import assert_array_equal

from ctdam.parser import CnvFile, CTDData
from ctdam.proc.modules import (
    AirPressureCorrection,
    AlignCTD,
    OwnBtlFile,
    WFilter,
    create_bottle_file,
    wildedit_geomar,
)
from ctdam.proc.utils import (
    BinnedDataError,
    get_alignment_delay_and_correlation_values,
)

logger = logging.getLogger(__name__)


@pytest.fixture
def cnv():
    return CnvFile(test_cnv, create_dataframe=False)


def test_align_get_corr(cnv):
    instance = AlignCTD()
    instance.parameters = cnv.parameters
    assert instance.get_correlation(cnv.parameters["sbeox0ML/L"]) == -1
    assert instance.get_correlation(cnv.parameters["c0mS/cm"]) == 1


def test_correlation_check():
    instance = AlignCTD()
    instance.minimum_correlation = 0.1
    instance.sample_rate = 24
    assert not instance.check_correlation_result(
        5 * instance.sample_rate, 0.09
    )
    assert not instance.check_correlation_result(
        6.001 * instance.sample_rate, 0.50
    )
    assert not instance.check_correlation_result(
        0.99 * instance.sample_rate, 0.50
    )
    assert instance.check_correlation_result(5.99 * instance.sample_rate, 0.50)
    assert not instance.check_correlation_result(np.nan, 0.50)
    assert not instance.check_correlation_result(
        2 * instance.sample_rate, np.nan
    )


def test_align_without_values(cnv):
    new_cnv = AlignCTD()(
        input=cnv,
        arguments={
            "Oxygen": None,
        },
    )
    assert isinstance(new_cnv, CTDData)
    try:
        assert_array_equal(
            CnvFile(
                test_cnv, create_dataframe=False
            ).parameters.get_full_data_array(),
            new_cnv.parameters.get_full_data_array(),
        )
    except AssertionError:
        if not new_cnv.processing_steps.get_step("alignctd"):
            assert False
        delay_values = get_alignment_delay_and_correlation_values(
            new_cnv.processing_steps
        )
        for delays in delay_values:
            assert 1 < float(delays[0]) < 6

    else:
        assert not new_cnv.processing_steps.get_step("alignctd")


def test_compare_align_with_and_without_values(cnv):
    without_values = AlignCTD()(
        input=cnv,
        arguments={
            "Oxygen": None,
        },
    )
    if not without_values.processing_steps.get_step("alignctd"):
        assert False
    delay_values = get_alignment_delay_and_correlation_values(
        without_values.processing_steps
    )
    with_values = AlignCTD()(
        input=CnvFile(test_cnv, create_dataframe=False),
        arguments={
            "Oxygen1": delay_values[0][0],
            "Oxygen2": delay_values[1][0],
        },
    )
    assert_array_equal(
        without_values.parameters.full_data_array,
        with_values.parameters.full_data_array,
    )


def test_wildedit_geomar_logic(cnv):
    data = cnv.parameters["t090C"].data
    new_data, flags = wildedit_geomar(
        data=data,
        flag=cnv.parameters["flag"].data,
        std1=3.0,
        std2=10,
        window_size=50,
    )
    assert new_data.size == data.size == flags.size


def test_alignment_factor_retrieval(cnv):
    result = AlignCTD()(
        input=cnv,
        arguments={
            "Oxygen": None,
        },
    )
    assert get_alignment_delay_and_correlation_values(
        result.processing_steps
    ) == [
        ("3.21", "0.29"),
        ("3.0", "0.14"),
    ]


def test_air_pressure_correction(cnv):
    old_data = cnv.parameters["prDM"].data
    new_cnv = AirPressureCorrection()(input=cnv)
    pressure_diff = (
        new_cnv.processing_steps.get_step("airpressurecorrection")
        .metadata["pressure_diff"]
        .split()[0]
    )
    assert old_data.size == new_cnv.parameters["prDM"].data.size
    assert float(old_data[0]) + float(pressure_diff) == float(
        new_cnv.parameters["prDM"].data[0]
    )


def test_airpressure_with_bugged_metadata(cnv):
    cnv.metadata["Air_Pressure"] = "definitely_not_a_number"
    new_cnv = AirPressureCorrection()(input=cnv)
    assert isinstance(new_cnv, CTDData)


def test_binned_data_error():
    cnv = cnv_path.joinpath("MSM140_1.cnv")
    with pytest.raises(BinnedDataError):
        AlignCTD()(
            input=cnv,
            arguments={
                "Oxygen": None,
            },
        )


def test_create_btl():
    btl = create_bottle_file(
        input=cnv_path.joinpath("EMB295_14-1.cnv"),
        arguments={
            "bl": btl_path.joinpath("EMB295_14-1.bl"),
            "write_btl": False,
        },
    )
    assert isinstance(btl, OwnBtlFile)


def test_wfilter(cnv):
    new_cnv = WFilter()(
        input=cnv,
        arguments={
            "tUrB iDIty": {"window_type": "median", "window_width": 10}
        },
    )
    assert isinstance(new_cnv, CTDData)
    # check for boundary effects
    assert new_cnv["prDM"].data[0] > 0.2
    assert cnv.parameters.get_data_length() == new_cnv.get_data_length()



from datetime import datetime
from pathlib import Path

import pytest
from conftest import btl_path, cnv_path, hex_path

from ctdam.parser import BottleFile, CnvFile, HexFile


@pytest.mark.parametrize(
    "file",
    [file for file in cnv_path.glob("*.cnv")],
)
class TestExampleFiles:
    @pytest.fixture
    def cnv(self, file: Path, run_long_tests: bool):
        if (not run_long_tests) and (file.stat().st_size > 5 * 1000000):
            pytest.skip(f"Skipping long cnv: {file}")
        try:
            cnv = CnvFile(
                file,
                create_dataframe=True,
            )
        except FileNotFoundError as error:
            pytest.skip(str(error))
        else:
            return cnv

    def test_found_start_time(self, cnv: CnvFile):
        assert isinstance(cnv.start_time, datetime)

    def test_calculate_full_time(self, cnv):
        assert (cnv.absolute_time_calculation()) == (
            len([p for p in cnv.parameters.keys() if p.startswith("time")]) > 0
        )

    def test_ctd_data2cnv_conversion(self, cnv):
        ctd_data = cnv.to_ctd_data()
        test_cnv = cnv.path_to_file.with_stem(
            f"conversion_test_{cnv.path_to_file.stem}"
        )
        ctd_data.to_cnv(test_cnv)
        # test seabird metadata
        assert cnv.sbe9_data == CnvFile(test_cnv).sbe9_data
        # test custom metadata
        assert cnv.metadata == CnvFile(test_cnv).metadata
        # test sensor xml
        assert cnv.sensor_data == CnvFile(test_cnv).sensor_data
        # test processing_steps
        assert cnv.processing_steps == CnvFile(test_cnv).processing_steps
        test_cnv.unlink()

    @pytest.mark.skip("Parameter needs a little rewrite.")
    def test_cnv_export_from_np_array(
        self,
        cnv: CnvFile,
    ):
        if cnv.parameters.duplicate_columns:
            assert True
        else:
            new_cnv_data = cnv.array2cnv()
            assert new_cnv_data == cnv.data

    def test_processing_step_extraction(self, cnv: CnvFile):
        assert cnv.processing_steps

    def test_sample_rate_retrieval(self, cnv):
        if cnv.file_name == "MSM140_1":
            target_rate = 1
        else:
            target_rate = 24
        assert cnv.parameters.sample_rate == target_rate


@pytest.mark.parametrize(
    "filename",
    [filename for filename in hex_path.glob("*.hex")],
)
class TestHexFiles:
    @pytest.fixture
    def hex(self, filename: str) -> HexFile:
        return HexFile(filename)

    def test_read_only_metadata(self, hex: HexFile):
        assert len(hex.data) == 0


@pytest.mark.parametrize(
    "filename",
    [filename for filename in btl_path.rglob("*.btl")],
)
def test_btl_reading(filename):
    btl = BottleFile(filename)
    assert "Timestamp" in btl.df.columns



import tkinter as tk
import tkinter.font as tkFont
from abc import abstractmethod
from pathlib import Path
from tkinter import filedialog
from typing import Callable

try:
    import customtkinter as ctk
    from CTkMessagebox import CTkMessagebox
except ImportError:
    raise ImportError(
        "The 'gui' extra is required to use this feature. Install with: pip install ctdam[gui]"
    )
import tomlkit
from tomlkit.toml_file import TOMLFile


class TomlEditor(ctk.CTkFrame):
    """General editor to configure toml files."""

    def __init__(
        self,
        master: ctk.CTkFrame,
        title: str,
        possible_parameters: list[str],
        config_file: Path | str = "",
        title_size: int = 20,
        fg_color: str = "transparent",
        default_dir_to_save_in: Path | str = "",
    ):
        super().__init__(master)
        self.master = master
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(2, weight=1)
        self.configure(fg_color=fg_color)

        self.module_frames = {}
        self.possible_parameters = possible_parameters
        self.default_dir_to_save_in = default_dir_to_save_in

        self.title = ctk.CTkLabel(
            self,
            text=title,
            font=(tkFont.nametofont("TkDefaultFont"), title_size),
        )
        self.title.grid(row=0, column=0, padx=15, pady=15)
        self.header_frame = ctk.CTkFrame(self)
        self.header_frame.configure(fg_color="transparent")
        self.header_frame.grid(row=1, column=0, sticky="ew")
        self.header_frame.grid_columnconfigure(0, weight=1)
        self.header_frame.grid_columnconfigure(1, weight=1)
        self.header_frame.grid_columnconfigure(2, weight=1)

        self.load_button = ctk.CTkButton(
            self.header_frame, text="Load TOML", command=self.load_toml
        )
        self.load_button.grid(row=0, column=0, padx=5, pady=5)

        self.save_button = ctk.CTkButton(
            self.header_frame, text="Save TOML", command=self.save_toml
        )
        self.save_button.grid(row=0, column=1, padx=5, pady=5)

        self.cancel_button = ctk.CTkButton(
            self.header_frame, text="Cancel", command=self.cancel
        )
        self.cancel_button.grid(row=0, column=2, padx=5, pady=5)

        self.content_frame = ctk.CTkScrollableFrame(self)
        self.content_frame.configure(fg_color="transparent")
        self.content_frame.grid(row=2, column=0, sticky="nsew")
        self.content_frame.bind_all(
            "<MouseWheel>",
            lambda e: self.content_frame._parent_canvas.yview(
                "scroll", -1, "units"
            ),
        )
        self.content_frame.bind_all(
            "<Button-4>",
            lambda e: self.content_frame._parent_canvas.yview(
                "scroll", -1, "units"
            ),
        )
        self.content_frame.bind_all(
            "<Button-5>",
            lambda e: self.content_frame._parent_canvas.yview(
                "scroll", 1, "units"
            ),
        )
        if config_file:
            self.load_config(config_file)

    def cancel(self):
        for frame in self.winfo_children():
            frame.destroy()
        self.destroy()
        self.master.destroy()

    def load_toml(self):
        file_path = filedialog.askopenfilename(
            filetypes=[("TOML Files", "*.toml")]
        )
        if file_path:
            self.file_path = Path(file_path).absolute()
            self.load_config(file_path)

    def save_toml(self):
        if not self.check_input():
            return
        file_path = filedialog.asksaveasfilename(
            defaultextension=".toml",
            filetypes=[("TOML Files", "*.toml")],
            initialfile=self.file_path.name,
            initialdir=(
                self.file_path.parent
                if not self.default_dir_to_save_in
                else self.default_dir_to_save_in
            ),
        )
        if file_path:
            try:
                with open(file_path, "w") as file:
                    file.write(tomlkit.dumps(self.config_data))
                CTkMessagebox(
                    title="Success",
                    message="Configuration saved successfully!",
                    icon="check",
                    option_1="Ok",
                )
                self.cancel()

            except Exception as e:
                print(e)
                CTkMessagebox(
                    title="Error",
                    message=f"Failed to save file: {e}",
                    icon="cancel",
                    option_1="Ok",
                )

    @abstractmethod
    def check_input(self) -> bool:
        pass

    def bad_input_warning(self, message: str):
        CTkMessagebox(
            title="Error",
            message=message,
            icon="cancel",
            option_1="Ok",
        )

    def load_config(self, file_path: Path | str):
        try:
            self.config_data = TOMLFile(file_path).read()
            self.file_path = Path(file_path).absolute()
            self.populate_fields()
        except Exception as e:
            CTkMessagebox(
                title="Error",
                message=f"Failed to load file: {e}",
                icon="cancel",
                option_1="Ok",
            )

    def populate_fields(self):
        for widget in self.content_frame.winfo_children():
            widget.destroy()
        self.module_frames.clear()
        row_index = 0
        for key in self.possible_parameters:
            if key in self.config_data:
                value = self.config_data[key]
            else:
                value = ""
            self.create_key_value_field(
                self.content_frame, key, value, row=row_index
            )
            row_index += 1

        self.load_config_specific_data(row_index)

    @abstractmethod
    def load_config_specific_data(self, row):
        pass

    def create_picker_element(
        self,
        frame: ctk.CTkFrame,
        entry: ctk.CTkEntry,
        directory: bool = True,
        callback: Callable | None = None,
        initial_dir: Path | str | None = None,
    ) -> ctk.CTkButton:
        if directory:
            command = filedialog.askdirectory
            text = "Pick directory"
            width = 80
        else:
            command = filedialog.askopenfilename
            text = "Pick file"
            width = 60

        def open_file_picker():
            entry.delete(0, tk.END)
            entry.insert(0, command(initialdir=initial_dir))
            entry.xview(tk.END)
            if isinstance(callback, Callable):
                callback()

        return ctk.CTkButton(
            frame, command=open_file_picker, text=text, width=width
        )

    def create_key_value_field(self, parent, key, value, row):
        frame = ctk.CTkFrame(parent, fg_color="transparent")
        frame.grid(row=row, column=0, sticky="ew", padx=5, pady=2)
        frame.grid_columnconfigure(0, weight=1)

        key_label = ctk.CTkLabel(frame, text=key, anchor="w")
        key_label.grid(row=0, column=0, sticky="w", padx=5, pady=2)

        entry = ctk.CTkEntry(frame, width=300)
        entry.grid(row=0, column=2)
        entry.insert(0, value)
        entry.xview(tk.END)

        def update_value(event=None):
            self.config_data[key] = entry.get()
            if not event:
                self.populate_fields()

        if "dir" in key:
            file_picker = self.create_picker_element(
                frame=frame,
                entry=entry,
                callback=update_value,
            )
            file_picker.grid(row=0, column=1, padx=5, pady=5)

        entry.bind("<FocusOut>", update_value)
        entry.bind("<Leave>", update_value)



from __future__ import annotations

import logging
import warnings
from collections import UserList
from pathlib import Path
from typing import Callable, Type

import numpy as np
import pandas as pd

from ctdam.parser import (
    BottleFile,
    BottleLogFile,
    CnvFile,
    DataFile,
    HexFile,
)
from ctdam.utils import get_unique_sensor_data

logger = logging.getLogger(__name__)


def get_collection(
    path_to_files: Path | str,
    file_suffix: str = "cnv",
    only_metadata: bool = False,
    pattern: str = "",
    sorting_key: Callable | None = None,
) -> Type[FileCollection]:
    """
    Factory to create instances of FileCollection, depending on input type.

    Parameters
    ----------
    path_to_files : Path | str :
        The path to the directory to search for files.
    file_suffix : str :
        The suffix to search for. (Default value = "cnv")
    only_metadata : bool :
        Whether to read only metadata. (Default value = False)
    pattern: str
        A filter for file selection. (Default value = '')
    sorting_key : Callable | None :
        A callable that returns the filename-part to use to sort the collection. (Default value = None)
    Returns
    -------
    An instance of FileCollection or one of its children.

    """
    mapping_suffix_to_type = {
        "cnv": CnvCollection,
        "btl": FileCollection,
        "bl": FileCollection,
        "hex": HexCollection,
    }
    file_suffix = file_suffix.strip(".")
    try:
        collection = mapping_suffix_to_type[file_suffix](
            path_to_files, file_suffix, only_metadata, pattern, sorting_key
        )
    except KeyError:
        raise ValueError(f"Unknown input file type: {file_suffix}, aborting.")
    else:
        return collection


class FileCollection(UserList):
    """
    A representation of multiple files of the same kind. These files share
    the same suffix and are otherwise closely connected to each other. A common
    use case would be the collection of CNVs to allow for easier processing or
    integration of field calibration measurements.

    Parameters
    ----------
    path_to_files : Path | str :
        The path to the directory to search for files.
    file_suffix : str :
        The suffix to search for. (Default value = "cnv")
    only_metadata : bool :
        Whether to read only metadata. (Default value = False)
    pattern: str
        A filter for file selection. (Default value = '')
    sorting_key : Callable | None :
        A callable that returns the filename-part to use to sort the collection. (Default value = None)
    """

    def __init__(
        self,
        path_to_files: str | Path,
        file_suffix: str,
        only_metadata: bool = False,
        pattern: str = "",
        sorting_key: Callable | None = None,
    ):
        super().__init__()
        self.path_to_files = Path(path_to_files)
        self.file_suffix = file_suffix.strip(".")
        self.pattern = pattern
        self.sorting_key = sorting_key
        self.file_type = self.extract_file_type(self.file_suffix)
        self.individual_file_paths = self.collect_files(
            pattern=pattern,
            sorting_key=sorting_key,
        )
        self.data = self.load_files(only_metadata)
        if not only_metadata:
            self.df_list = self.get_dataframes()
            self.df = self.get_collection_dataframe(self.df_list)

    def __str__(self):
        return "/n".join(self.data)

    def extract_file_type(self, suffix: str) -> Type[DataFile]:
        """
        Determines the file type using the input suffix.

        Parameters
        ----------
        suffix : str :
            The file suffix.
        Returns
        -------
        An object corresponding to the given suffix.
        """
        mapping_suffix_to_type = {
            "cnv": CnvFile,
            "btl": BottleFile,
            "bl": BottleLogFile,
            "hex": HexFile,
        }
        file_type = DataFile
        for key, value in mapping_suffix_to_type.items():
            if key == suffix:
                file_type = value
                break
        return file_type

    def collect_files(
        self,
        pattern: str = "",
        sorting_key: Callable | None = lambda file: int(
            file.stem.split("_")[3]
        ),
    ) -> list[Path]:
        """
        Creates a list of target files, recursively from the given directory.
        These can be sorted with the help of the sorting_key parameter, which
        is a Callable that identifies the part of the filename that shall be
        used for sorting.

        Parameters
        ----------
        pattern: str
            A filter for file selection. Is given to rglob. (Default value = '')
        sorting_key : Callable | None :
            The part of the filename to use in sorting. (Default value = lambda file: int(file.stem.split("_")[3]))
        Returns
        -------
        A list of all paths found.
        """
        if self.path_to_files.is_file():
            return [self.path_to_files]
        else:
            return sorted(
                self.path_to_files.rglob(f"*{pattern}*{self.file_suffix}"),
                key=sorting_key,
            )

    def load_files(self, only_metadata: bool = False) -> list[DataFile]:
        """
        Creates python instances of each file.

        Parameters
        ----------
        only_metadata : bool :
            Whether to load only file metadata. (Default value = False)
        Returns
        -------
        A list of all instances.
        """
        data = []
        for file in self.individual_file_paths:
            try:
                data.append(
                    self.file_type(
                        path_to_file=file,
                        only_header=only_metadata,
                    )
                )
            except TypeError:
                logger.error(
                    f"Could not open file {file} with the type "
                    f"{self.file_type}."
                )
                continue
        return data

    def get_dataframes(
        self,
        event_log: bool = False,
        coordinates: bool = False,
        time_correction: bool = False,
        cast_identifier: bool = False,
    ) -> list[pd.DataFrame]:
        """
        Collects all individual dataframes and allows additional column
        creation.

        Parameters
        ----------
        event_log : bool :
            (Default value = False)
        coordinates : bool :
            (Default value = False)
        time_correction : bool :
            (Default value = False)
        cast_identifier : bool :
            (Default value = False)

        Returns
        -------
        A list of the individual pandas DataFrames.
        """
        for index, file in enumerate(self.data):
            if event_log:
                file.add_station_and_event_column()
            if coordinates:
                file.add_position_columns()
            if time_correction:
                file.absolute_time_calculation()
                file.add_start_time()
            if cast_identifier:
                file.add_cast_number(index + 1)
        return [file.df for file in self.data]

    def get_collection_dataframe(
        self, list_of_dfs: list[pd.DataFrame] | None = None
    ) -> pd.DataFrame:
        """
        Creates one DataFrame from the individual ones, by concatenation.

        Parameters
        ----------
        list_of_dfs : list[pd.DataFrame] | None :
            A list of the individual DataFrames. (Default value = None)
        Returns
        -------
        A pandas DataFrame representing the whole dataset.
        """
        if not list_of_dfs:
            list_of_dfs = self.get_dataframes()
        if not list_of_dfs:
            raise ValueError("No dataframes to concatenate.")
        df = pd.concat(list_of_dfs, ignore_index=True)
        self.df = df
        return df

    def tidy_collection_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Apply the different dataframe edits to the given dataframe.

        Parameters
        ----------
        df : pd.DataFrame :
            A DataFrame to edit.
        Returns
        -------
        The tidied dataframe.
        """
        df = self.use_bad_flag_for_nan(df)
        df = self.set_dtype_to_float(df)
        return self.select_real_scan_data(df)

    def use_bad_flag_for_nan(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Replace all Nan values by the bad flag value, defined inside the files.

        Parameters
        ----------
        df : pd.DataFrame :
            The dataframe to edit.
        Returns
        -------
        The edited DataFrame.
        """
        bad_flags = set()
        for file in self.data:
            for line in file.data_table_description:
                if line.startswith("bad_flag"):
                    flag = line.split("=")[1].strip()
                    bad_flags.add(flag)
        for flag in bad_flags:
            df.replace(to_replace=flag, value=np.nan, inplace=True)
        return df

    def set_dtype_to_float(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Use the float-dtype for all DataFrame columns.

        Parameters
        ----------
        df : pd.DataFrame :
            The dataframe to edit.
        Returns
        -------
        The edited DataFrame.
        """
        for parameter in df.columns:
            if parameter in ["datetime"]:
                continue
            df[parameter] = df[parameter].astype("float")
            continue
        return df

    def select_real_scan_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Drop data rows have no 'Scan' value, if that column exists.

        Parameters
        ----------
        df : pd.DataFrame :
            The dataframe to edit.
        Returns
        -------
        The edited DataFrame.
        """
        try:
            scan_column = [
                c for c in df.columns if c.lower().startswith("scan")
            ][0]
        except IndexError:
            return df
        else:
            df = df.loc[df[scan_column].notna()]
        return df

    def to_csv(self, file_name):
        """
        Writes a csv file with the given filename.

        Parameters
        ----------
        file_name :
            The new csv file name.
        """
        self.df.to_csv(file_name)


class CnvCollection(FileCollection):
    """
    Specific methods to work with collections of .cnv files.
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        if len(args) < 3 and "file_suffix" not in kwargs:
            kwargs["file_suffix"] = "cnv"
        super().__init__(*args, **kwargs)
        self.data_meta_info = self.get_data_table_meta_info()
        self.sensor_data = get_unique_sensor_data(
            [file.sensors for file in self.data]
        )
        self.array = self.get_array()
        self.processing_steps = self.get_processing_steps()

    def get_dataframes(
        self,
        event_log: bool = False,
        coordinates: bool = False,
        time_correction: bool = False,
        cast_identifier: bool = False,
    ) -> list[pd.DataFrame]:
        """
        Collects all individual dataframes and allows additional column
        creation.

        Parameters
        ----------
        event_log : bool :
            (Default value = False)
        coordinates : bool :
            (Default value = False)
        time_correction : bool :
            (Default value = False)
        cast_identifier : bool :
            (Default value = False)
        Returns
        -------
        A list of the individual pandas DataFrames.
        """
        for index, file in enumerate(self.data):
            if event_log:
                file.add_station_and_event_column()
            if coordinates:
                file.add_position_columns()
            if time_correction:
                file.absolute_time_calculation()
                file.add_start_time()
            if cast_identifier:
                file.add_cast_number(index + 1)
        return [file.create_dataframe() for file in self.data]

    def get_data_table_meta_info(self) -> list[dict]:
        """
        Ensures the same data description in all input cnv files and returns
        it.
        Acts as an early alarm when working on different kinds of files, which
        cannot be concatenated together.

        Returns
        -------
        A list of dictionaries that represent the data column information.
        """
        all_column_descriptions = [
            file.parameters.get_metadata() for file in self.data
        ]
        for index, info in enumerate(all_column_descriptions):
            if all_column_descriptions[0] != info:
                for expected, real in zip(
                    all_column_descriptions[0].items(), info.items()
                ):
                    # allow difference in latitude inside depth
                    if expected[0] == "depSM":
                        if real[0] != "depSM":
                            raise AssertionError(
                                f"Data files {self.data[0].path_to_file} and {self.data[index].path_to_file} differ in:\n{expected} and {real}"
                            )

                    elif expected != real:
                        warnings.warn(
                            f"Data files {self.data[0].path_to_file} and {self.data[index].path_to_file} differ in:\n{expected} and {real}"
                        )

        return all_column_descriptions[0]

    def get_array(self) -> np.ndarray:
        """
        Creates a collection array of all individual file arrays.

        Returns
        -------
        A numpy array, representing the data of all input files.
        """
        return np.concatenate(
            [file.parameters.get_full_data_array() for file in self.data]
        )

    def get_processing_steps(self) -> list:
        """
        Checks the processing steps in the different files for consistency.
        Returns the steps of the first file, which should be the same as for
        all other files.

        Returns
        -------
        A list of ProcessingSteps.
        """
        individual_processing_steps = [
            file.processing_steps.modules for file in self.data
        ]
        for index, step_info in enumerate(individual_processing_steps):
            if step_info != individual_processing_steps[0]:
                message = f"The processing steps conducted on these files differ. First occurence between index 0 and {index}."
                warnings.warn(message)
                logger.warning(message)
        return individual_processing_steps[0]


class HexCollection(FileCollection):
    """
    Specific methods to work with collections of .hex files.

    Especially concerned with the detection of corresponding .XMLCON files.
    """

    def __init__(
        self,
        *args,
        xmlcon_pattern: str = "",
        path_to_xmlcons: Path | str = "",
        **kwargs,
    ):
        if len(args) < 3 and "file_suffix" not in kwargs:
            kwargs["file_suffix"] = "hex"
        # force only_metadata, as the hex data cannot be put into a DataFrame
        kwargs["only_metadata"] = True
        super().__init__(*args, **kwargs)
        if not xmlcon_pattern:
            xmlcon_pattern = self.pattern
        self.xmlcon_pattern = xmlcon_pattern
        self.path_to_xmlcons = (
            Path(path_to_xmlcons)
            if path_to_xmlcons
            else self.path_to_files.parent
        )
        self.xmlcons = self.get_xmlcons()

    def get_xmlcons(self) -> list[str]:
        """
        Returns all .xmlcon files found inside the root directory and its
        children, matching a given pattern.

        Does use the global sorting_key to attempt to also sort the xmlcons the
        same way.
        This is meant to be used in the future for a more specific hex-xmlcon
        matching.

        Returns
        -------
        A list of the found xmlcon filenames.
        """
        try:
            xmlcons = [
                Path(xmlcon_path).stem
                for xmlcon_path in sorted(
                    self.path_to_xmlcons.rglob(
                        f"*{self.xmlcon_pattern}*.XMLCON"
                    ),
                    key=self.sorting_key,
                )
            ]
        except (KeyError, IndexError):
            xmlcons = [
                Path(xmlcon_path).stem
                for xmlcon_path in self.path_to_xmlcons.rglob(
                    f"*{self.xmlcon_pattern}*.XMLCON"
                )
            ]
        return xmlcons



import json
import logging
import random
import webbrowser
from pathlib import Path

from bokeh.models import (
    ColumnDataSource,
    CustomJS,
    HoverTool,
    LinearAxis,
    Range1d,
    Title,
)
from bokeh.plotting import figure, output_file, save, show
from bs4 import BeautifulSoup
from tomlkit.toml_file import TOMLFile

from ctdam.conv.hexdecoder import decode_hex
from ctdam.parser import CnvFile, CTDData

logger = logging.getLogger(__name__)


def check_and_create_path(dir: Path | str):
    dir = Path(dir)
    if dir == Path("."):
        return

    if not dir.exists():
        dir.mkdir(parents=True)


def cruise_plots(
    directory: Path | str = "",
    output_directory: Path | str = "html",
    output_name: str = "main.html",
    embed_contents: bool = False,
    html_title: str = "",
    overwrite: bool = False,
    no_new_plots: bool = False,
    size_limit: int = 10,
    filter: str = "",
    show_html: bool = True,
    config_path: Path | str = "vis_config.toml",
    file_type: str = "cnv",
) -> Path | None:
    if not no_new_plots:
        output_directory = (
            Path(output_directory)
            if str(output_directory)
            else Path(directory)
        )
        if not output_directory.exists():
            output_directory.mkdir()
        if not file_type:
            file_type = ".cnv"

        file_type = f".{file_type}" if not file_type[0] == "." else file_type

        for file in Path(directory).glob(f"*{filter}*{file_type}"):
            if file.stat().st_size > size_limit * 1000000:
                logger.info(f"{file} above size limit of {size_limit}MB")
                continue
            if (
                Path(output_directory)
                .joinpath(file.name)
                .with_suffix(".html")
                .exists()
            ) and not overwrite:
                continue
            try:
                basic_bokeh_plot(
                    ctd_data=str(file),
                    output_directory=output_directory,
                    print_plot=True,
                    metadata=True,
                    show_plot=False,
                    config_path=config_path,
                )
            except Exception as error:
                logger.warning(f"Could not create a plot for {file}: {error}")
                continue

    if output_directory:
        directory = output_directory

    output_path = create_main_html(
        directory_path=directory,
        output_name=output_name,
        output_directory=output_directory,
        embed_contents=embed_contents,
        title=html_title,
        show_html=show_html,
    )
    return output_path


def basic_bokeh_plot(
    ctd_data: CTDData | CnvFile | Path | str,
    print_plot: bool = False,
    output_name: str = "",
    output_directory: Path | str = "",
    metadata: bool = True,
    show_plot: bool = True,
    y_axis_params: list[str] = ["prDM", "depSM"],
    config_path: Path | str = "vis_config.toml",
):
    if isinstance(ctd_data, Path | str):
        suffix = Path(ctd_data).suffix
        if suffix == ".cnv":
            ctd_data = CnvFile(ctd_data)
        elif suffix == ".hex":
            ctd_data = decode_hex(ctd_data)

    try:
        file_path = ctd_data.metadata_source.path_to_file
    except AttributeError:
        file_path = ctd_data.path_to_file

    source = ColumnDataSource(ctd_data.parameters.get_pandas_dataframe())

    try:
        config = TOMLFile(config_path).read()
    except Exception:
        try:
            config = TOMLFile(
                Path(__file__).parent.joinpath(config_path)
            ).read()
        except Exception:
            config = {}

    y_axis_param = ""
    y_axis_label = ""
    for param in y_axis_params:
        for p in ctd_data.parameters.get_parameter_list():
            if param == p.name:
                y_axis_param = param
                y_axis_label = p.metadata["longinfo"]
                break

    if not y_axis_param:
        logger.info(
            f"Could not find any of {y_axis_params} inside {file_path}"
        )
        return

    p = figure(
        y_axis_label=y_axis_label,
        sizing_mode="stretch_both",
        tools="pan, box_zoom, wheel_zoom, xwheel_zoom, ywheel_zoom, reset, save",
        active_drag="pan",
        active_scroll="wheel_zoom",
    )
    p.xaxis.visible = False
    non_plotting = [
        "flag",
        "dz/dtM",
        "timeS",
        "scan",
        "nbf",
        "nbin",
        "latitude",
        "longitude",
        "altM",
    ] + y_axis_params

    parameters = [
        param
        for param in ctd_data.parameters.get_parameter_list()
        if param.name not in non_plotting
    ]

    p.extra_x_ranges = {
        param.name: Range1d(start=0, end=param.span[1]) for param in parameters
    }

    p.y_range = Range1d(
        start=ctd_data.parameters[y_axis_param].span[1],
        end=ctd_data.parameters[y_axis_param].span[0],
    )

    colors = [
        f"#{random.randint(0, 0xFFFFFF):06x}" for _ in range(len(parameters))
    ]

    if metadata:
        title = Title(
            text=" | ".join(
                [f"{k} = {v}" for k, v in ctd_data.metadata.items()]
            ),
            text_font_size="8pt",
            align="left",
            text_color="black",
        )
        p.add_layout(title, "above")
        proc_meta = Title(
            text="".join(ctd_data.processing_steps._form_processing_info()),
            text_font_size="0pt",
            align="center",
            text_color="gray",
        )
        p.add_layout(proc_meta, "below")

    for index, parameter in enumerate(parameters):
        color = colors[index]
        name = parameter.name
        param_type = parameter.param.lower()
        unit = parameter.unit
        label = f"{name} [{unit}]"
        show_param = None

        def _use_config_data(info_dict):
            sensor = parameter.sensor_number - 1
            try:
                color = info_dict["colors"][sensor]
            except KeyError:
                color = colors[index]
            try:
                p.extra_x_ranges[name] = Range1d(
                    start=info_dict["span_start"],
                    end=info_dict["span_end"],
                )
            except KeyError:
                pass
            try:
                show_param = bool(info_dict["show"])
            except KeyError:
                show_param = None
            return color, show_param

        if config:
            matches = [p for p in config if param_type.startswith(p)]
            for match in matches:
                for unit_desc in config[match]:
                    if (
                        unit_desc.replace("-", " ").replace("_", "/")
                        in unit.lower()
                    ):
                        color, show_param = _use_config_data(
                            config[match][unit_desc]
                        )
                        break
                    elif unit_desc in [
                        "show",
                        "colors",
                        "span_start",
                        "span_end",
                    ]:
                        color, show_param = _use_config_data(config[match])
                        break

        xaxis = LinearAxis(
            x_range_name=name,
            axis_label_text_color=color,
            major_label_text_color=color,
            major_tick_line_color=color,
            axis_line_color=color,
        )

        line = p.line(
            name,
            y_axis_param,
            source=source,
            line_width=2,
            legend_label=label,
            color=color,
            x_range_name=name,
        )
        p.add_layout(xaxis, "below")

        p.add_tools(
            HoverTool(
                renderers=[line],
                tooltips=[
                    ("Name", label),
                    (
                        "Color",
                        '<span class="bk-tooltip-color-block" '
                        'style="background-color:{}"> </span>'.format(color),
                    ),
                    (f"{unit}", "$x"),
                    ("db", "$y"),
                ],
            )
        )

        xaxis.visible = line.visible = _auto_show_plot(name, unit, show_param)

        # JavaScript callback to hide/show the x-axis when the line is hidden/shown
        callback = CustomJS(
            args=dict(xaxis=xaxis, line=line),
            code="""
            xaxis.visible = line.visible;
        """,
        )
        line.js_on_change("visible", callback)

    p.legend.location = "top_left"
    p.legend.click_policy = "hide"
    p.legend.background_fill_alpha = 0.1
    p.legend.background_fill_color = None

    if print_plot:
        output_name = file_path.stem if not output_name else str(output_name)
        output_directory = (
            Path(output_directory).absolute()
            if output_directory
            else file_path.parent.absolute()
        )
        check_and_create_path(output_directory)
        output_file(
            output_directory.joinpath(output_name).with_suffix(".html"),
            title=f"Plot of {file_path.name}",
        )
        save(p)

    if show_plot:
        show(p)


def _auto_show_plot(name: str, unit: str, show_param: bool | None) -> bool:
    if isinstance(show_param, bool):
        return show_param
    # Temperature
    if "deg C" in unit and not name.startswith(("pta", "potemp")):
        return True
    # Salinity
    elif "PSU" in unit:
        return True
    # Oxygen
    elif name.startswith("sb") and "%" not in unit:
        return True
    else:
        return False


def create_main_html(
    directory_path: Path | str,
    output_name: str = "main_plots.html",
    output_directory: Path | str = "",
    embed_contents: bool = True,
    title: str = "",
    show_html: bool = True,
) -> Path | None:
    check_and_create_path(directory_path)
    check_and_create_path(output_directory)
    html_files = [
        file
        for file in Path(directory_path).iterdir()
        if file.suffix == ".html"
    ]
    if not html_files:
        print(f"No HTML files found in {directory_path}")
        return

    dropdown_options = []
    plot_iframes = []
    plot_metadata = {}

    for i, html_file in enumerate(sorted(html_files)):
        if html_file.name == output_name or html_file.name.startswith("."):
            continue
        plot_id = f"plot_{i}"

        # extract metadata from the Bokeh JSON data in the HTML file
        with open(html_file, "r", encoding="utf-8") as f:
            soup = BeautifulSoup(f.read(), "html.parser")
            script_tags = soup.find_all("script", {"type": "application/json"})

            metadata = {"title": html_file.stem, "text": ""}
            for script in script_tags:
                try:
                    # custom metadata header
                    data = json.loads(script.string)
                    html_metadata = data[list(data.keys())[0]]["roots"][0][
                        "attributes"
                    ]["above"][0]["attributes"]["text"]
                    metadata["text"] = "\n".join(html_metadata.split(" | "))
                except (json.JSONDecodeError, KeyError, TypeError):
                    pass
                try:
                    # ctdam.proc module info
                    data = json.loads(script.string)
                    html_metadata = data[list(data.keys())[0]]["roots"][0][
                        "attributes"
                    ]["below"][1]["attributes"]["text"]
                    metadata["ctdam.proc"] = "".join(html_metadata)
                except Exception:
                    pass

        plot_metadata[plot_id] = metadata
        dropdown_options.append(
            f'<option value="{plot_id}">{metadata["title"]}</option>'
        )
        if embed_contents:
            with open(html_file, "r", encoding="utf-8") as f:
                html_content = f.read()
            plot_iframes.append(
                f'<div id="{plot_id}_content" style="display: none;">{html_content}</div>'
            )
        else:
            plot_iframes.append(
                f'<iframe id="{plot_id}_content" src="{Path(html_file).absolute()}" style="display: none; border: none;"></iframe>'
            )

    dropdown_options_html = "\n".join(dropdown_options)
    plot_iframes_html = "\n".join(plot_iframes)
    title = f"{directory_path} Plots" if not title else title

    main_html = f"""
    <!DOCTYPE html>
    <html>
    <head>
        <title>{title}</title>
        <style>
            body {{
                font-family: Arial, sans-serif;
                margin: 0;
                padding: 0;
                display: flex;
                flex-direction: column;
                height: 100vh;
            }}
            h1 {{
                text-align: center;
                background-color: #f0f0f0;
                padding: 10px;
                margin: 0;
            }}
            .container {{
                display: flex;
                flex: 1;
                overflow: hidden;
            }}
            .sidebar {{
                width: 250px;
                padding: 10px;
                background-color: #f9f9f9;
                border-right: 1px solid #ddd;
                overflow-y: auto;
            }}
            .plot-display-container {{
                flex: 1;
                padding: 10px;
                overflow-y: auto;
                position: relative;
                height: calc(100vh - 100px);
            }}
            .plot-content, .plot-iframe {{
                position: absolute;
                top: 0;
                left: 0;
                width: 100%;
                // height: 100%;
                height: calc(100vh - 180px);
                border: none;
            }}
            .metadata {{
                margin-top: 20px;
                padding: 10px;
                background-color: #f0f0f0;
                border-radius: 5px;
                white-space: pre-wrap;
            }}
            select {{
                width: 100%;
                padding: 8px;
                margin-bottom: 10px;
            }}
            .proc_meta {{
                font-size: 12px;
            }}

            @media print {{
                body {{
                    visibility: hidden;
                    overflow: visible !important;
                }}
                .sidebar {{
                    display: none;
                }}
                .container {{
                    visibility: visible;
                    position: absolute;
                    top: 0;
                    left: 0;
                    height: 100vh;
                    width: 100vw;
                    overflow: visible !important;
                }}
            }}
            @page {{
                size: auto;
                margin: 0;
            }}
        </style>
        <script>
            // Embed the metadata as a JavaScript object
            const plotMetadata = {json.dumps(plot_metadata)};

            function updatePlot() {{
                const dropdown = document.getElementById("plot_dropdown");
                const selectedPlotId = dropdown.value;
                const plotDisplayContainer = document.querySelector('.plot-display-container');

                // Hide all plots
                document.querySelectorAll('[id$="_content"]').forEach(plot => {{
                    plot.style.display = "none";
                }});

                // Show the selected plot
                const selectedPlot = document.getElementById(selectedPlotId + "_content");
                if (selectedPlot) {{
                    selectedPlot.style.display = "block";
                    selectedPlot.style.position = "absolute";
                    selectedPlot.style.top = "0";
                    selectedPlot.style.left = "0";
                    selectedPlot.style.width = "100%";
                    selectedPlot.style.height = "100%";
                }}

                // Update metadata
                const metadataDiv = document.getElementById("metadata");
                if (selectedPlotId) {{
                    const selectedMetadata = plotMetadata["{'" + selectedPlotId + "'}"];

                    metadataDiv.innerHTML = `<h3>Metadata</h3><p>${{selectedMetadata.text}}</p><h4>Processing Module Info</h4><p id="proc_meta" class="proc_meta">${{selectedMetadata.ctdam.proc}}</p>`;
                }} else {{
                    metadataDiv.innerHTML = `<p>Select a plot to view metadata.</p>`;
                }}
            }}

            function printSelectedPlot() {{
                const selectedPlotId = document.getElementById('plot_dropdown').value;
                if (!selectedPlotId) {{
                    alert('Please select a plot to print.');
                    return;
                }}

                const selectedPlot = document.getElementById(selectedPlotId + "_content");
                if (selectedPlot) {{
                    selectedPlot.style.display = "block";
                }}

                window.print();
        }}
        </script>
    </head>
    <body>
        <h1>{title}</h1>
        <div class="container">
            <div class="sidebar">
                <label for="plot_dropdown">Select Plot:</label>
                <select id="plot_dropdown" onchange="updatePlot()" multiple=True size=5 autofocus=True>
                    {dropdown_options_html}
                </select>
                <button id="print-button" onclick="printSelectedPlot()">Print</button>
                <div id="metadata" class="metadata">
                    <p>Select a plot to view metadata.</p>
                </div>
            </div>
            <div class="plot-display-container">
                {plot_iframes_html}
            </div>
        </div>
    </body>
    </html>
    """

    output_directory = (
        Path(output_directory) if output_directory else Path(directory_path)
    )
    output_path = output_directory.joinpath(output_name)
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(main_html)
    if show_html:
        webbrowser.open_new_tab(f"file://{output_path.absolute()}")
    return output_path



import logging

import numpy as np
import pytest
from conftest import test_cnv

from ctdam.parser import CnvFile, Parameters

logger = logging.getLogger(__name__)

# TODO: rethink this
data_table_length = 3067


@pytest.fixture
def cnv():
    return CnvFile(test_cnv)


@pytest.fixture
def parameters(cnv: CnvFile):
    return Parameters(cnv.data, cnv.data_table_description)


def test_parameter_metadata_loading(parameters: Parameters):
    spans = parameters.get_spans()
    assert isinstance(spans, list)
    for parameter in parameters.values():
        assert isinstance(parameter.metadata, dict)
        assert isinstance(parameter.span, tuple)
        parameter.data[0] = parameter.span[1] + 1
    assert spans != parameters.get_spans()


@pytest.mark.parametrize(
    "data, metadata, name",
    [
        (
            np.full(fill_value=1, shape=data_table_length, dtype=float),
            {},
            "something",
        ),
        ("EMB999_123-12", {"shortname": "event", "name": "event"}, "event"),
    ],
)
def test_addition_of_new_parameters(
    parameters: Parameters,
    data: np.ndarray | str | int | float,
    metadata: dict,
    name: str,
):
    new_parameter = parameters.create_parameter(
        data=data,
        metadata=metadata,
        name=name,
    )
    assert new_parameter.name in parameters
    assert new_parameter.data.shape[0] == data_table_length
    assert len(new_parameter.metadata) == 5



import logging
import re
from datetime import datetime

import pandas as pd

from ctdam.parser import DataFile

logger = logging.getLogger(__name__)


class BottleLogFile(DataFile):
    """Bottle Log file representation, that extracts the three different data
    types from the file: reset time and the table with bottle IDs and
    corresponding data ranges.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, path_to_file, create_dataframe=False):
        super().__init__(path_to_file)
        self.reset_time = self.obtaining_reset_time()
        self.origin_cnv = self.raw_file_data[0].strip()
        self.data = self.data_whitespace_removal()

        if create_dataframe:
            self.original_df = self.create_dataframe()
            self.df = self.original_df
        else:
            self.data_list = self.create_list()

    def data_whitespace_removal(self) -> list:
        """Strips the input from whitespace characters, in this case especially
        newline characters.

        Parameters
        ----------

        Returns
        -------
        the original data stripped off the whitespaces

        """
        temp_data = []
        for line in self.raw_file_data[2:]:
            temp_data.append(line.strip())
        return temp_data

    def obtaining_reset_time(self) -> datetime:
        """Reading reset time with small input check.

        Parameters
        ----------

        Returns
        -------
        a datetime.datetime object of the device reset time

        """

        regex_check = re.search(
            r"RESET\s(\w{3}\s\d+\s\d{4}\s\d\d:\d\d:\d\d)",
            self.raw_file_data[1],
        )
        if regex_check:
            return datetime.strptime(regex_check.group(1), "%b %d %Y %H:%M:%S")
        else:
            error_message = """BottleLogFile is not formatted as expected:
                Reset time could not be extracted."""
            logger.error(error_message)
            raise IOError(error_message)

    def create_list(self) -> list:
        """Creates a list of usable data from the list specified in self.data.
        the list consists of: an array of ID's representing the bottles, the date and time of the data sample
        and the lines of the cnv corresponding to the bottles

        Parameters
        ----------

        Returns
        -------
        a list representing the bl files table information
        """
        content_array = []
        for i in range(len(self.data)):
            bottles = [int(x) for x in self.data[i].split(",")[:2]]
            date = self.convert_date(self.data[i].split(",")[2])
            lines = tuple([int(x) for x in self.data[i].split(",")[3:]])

            content_array.append([bottles, date, lines])

        return content_array

    def convert_date(self, date: str):
        """Converts the Dates of the .bl files to an ISO 8601 standard

        Parameters
        ----------

        Returns
        -------
        a string with the date in the form of "yymmddThhmmss"
        """
        date = date.strip()
        month_list = [
            "Jan",
            "Feb",
            "Mar",
            "Apr",
            "May",
            "Jun",
            "Jul",
            "Aug",
            "Sep",
            "Oct",
            "Nov",
            "Dec",
        ]

        month_ind = month_list.index(date.split(" ")[0]) + 1
        if month_ind < 10:
            month = "0" + str(month_ind)
        else:
            month = str(month_ind)
        day = date.split(" ")[1]
        year = (date.split(" ")[2])[2:]
        time = date.split(" ")[3].replace(":", "")
        return year + month + day + "T" + time

    def create_dataframe(self) -> pd.DataFrame:
        """Creates a dataframe from the list specified in self.data.

        Parameters
        ----------

        Returns
        -------
        a pandas.Dataframe representing the bl files table information
        """
        data_lists = []
        for line in self.data:
            inner_list = line.split(",")
            # dropping first column as its the index
            data_lists.append(inner_list[1:])
        df = pd.DataFrame(data_lists)
        df.columns = ["Bottle ID", "Datetime", "start_range", "end_range"]
        return df



import logging
import re
from itertools import zip_longest
from typing import Tuple

logger = logging.getLogger(__name__)


def read_event_name(
    station_string: str,
    regex_string: str = r"(?P<c>[a-z]{1,3}\d{1,3})(-|_|\/)?(?P<cn>1|2)?(-|_)(?P<s>\d{1,4})(-|_)(?P<e>\d{1,2})",
) -> Tuple[str, str]:
    match = re.match(regex_string, station_string, flags=re.I)
    if match:
        match_dict = dict(match.groupdict())
        cruise_name = match_dict["c"]
        if "cn" in match_dict:
            if match_dict["cn"]:
                cruise_name = f"{match_dict['c']}_{match_dict['cn']}"

        if "s" in match_dict:
            station = match_dict["s"]
        else:
            station = ""
        if "e" in match_dict:
            event = match_dict["e"]
        else:
            event = ""
    else:
        cruise_name = station = event = ""
    if station:
        station = f"{station}-{event}"
    else:
        station = ""
    return cruise_name, station


def create_event_string(
    cruise: str,
    station_event: str,
    leading_zeroes: bool = True,
) -> str:
    if station_event == "":
        return ""
    station, event = station_event.split("-")
    if leading_zeroes:
        return f"{cruise.upper()}_{float(station):03.0f}-{int(event):02.0f}"
    else:
        return f"{cruise.upper()}_{int(station)}-{int(event)}"


def parse_xmlcon_sensor_data(sensor_info: dict) -> dict:
    sensor_info = sensor_info["SBE_InstrumentConfiguration"]["Instrument"][
        "SensorArray"
    ]
    # rename sensor array size -> count
    sensor_info = {
        "@count" if k == "@Size" else k: v for k, v in sensor_info.items()
    }
    # rename Sensor -> sensor
    sensor_info = {
        "sensor" if k == "Sensor" else k: v for k, v in sensor_info.items()
    }
    for sensor in sensor_info["sensor"]:
        # remove redudant SensorID
        sensor.pop("@SensorID")
        # rename index -> Channel
        sensor["@Channel"] = str(int(sensor.pop("@index")) + 1)

    return sensor_info


def extract_sensor_name(sensors: dict) -> list:
    # create a tidied version of the xml-parsed sensor dict
    sensor_names = []
    tidied_sensor_list = []
    for entry in sensors:
        try:
            sensor_key = [
                key
                for key in entry.keys()
                if key.endswith(("Sensor", "Meter"))
            ][0]
        except IndexError:
            continue
        sensor_name = sensor_key.removesuffix("Sensor")
        # the wetlab sensors feature a suffix _Sensor
        sensor_name = sensor_name.removesuffix("_")
        # assuming, that the first sensor in the xmlcon is also on the
        # first sensor strand, the second occurence of the name is
        # suffixed with '2'
        if sensor_name in sensor_names:
            sensor_name += "2"
        sensor_names.append(sensor_name)
        # move the calibration info one dictionary level up
        calibration_info = entry[sensor_key]
        # build the new dictionary
        try:
            new_dict = {
                "Channel": str(int(entry["@index"]) + 1),
                "SensorName": sensor_name,
                **calibration_info,
            }
        except Exception:
            new_dict = {
                "Channel": entry["@Channel"],
                "SensorName": sensor_name,
                **calibration_info,
            }
        tidied_sensor_list.append(new_dict)
    return tidied_sensor_list


def get_unique_sensor_data(
    sensor_data: list[list[dict]],
) -> list[tuple[list[dict]]]:
    """
    Returns all the unique sensors and their configuration used in the given
    collection of sensor data. These will typically be parsed from xml inside
    .cnv or .xmlcon files.
    If for example, the first oxygen sensor has been replaced after the 8 cast,
    then we will see that in the output structure by a seconde tuple, with the
    number 8 and the individual sensor information for that new oxygen sensor.

    Parameters
    ----------
    sensor_data:
        The structure of xml-parsed dicts inside two organizing lists.

    Returns
    -------
    The input structure stripped down to unique sensor data and appended by
    the index, at which this new sensor appeared the first time.

    """
    unique = []
    last_unique = None
    for index, individual_sensor_data in enumerate(
        [file for file in sensor_data]
    ):
        if last_unique is None:
            unique.append((index, individual_sensor_data))
        else:
            differing_dicts = [
                current_dict
                for last_dict, current_dict in zip_longest(
                    last_unique, individual_sensor_data
                )
                if current_dict != last_dict
            ]
            if differing_dicts:
                unique.append((index, differing_dicts))
        last_unique = individual_sensor_data
    return unique


class UnexpectedFileFormat(Exception):
    def __init__(self, file_type: str, error: str) -> None:
        message = f"{file_type} is not formatted as expected: {error}"
        logger.error(message)
        super().__init__(message)



from pathlib import Path

import gsw
from seabirdscientific import processing as sbs_proc

from ctdam.proc.module import Module
from ctdam.proc.modules.air_pressure_correction import AirPressureCorrection
from ctdam.proc.modules.create_bottlefile import create_bottle_file
from ctdam.proc.modules.external_functions import (
    ExternalFunctionCaller,
    ExternalFunctions,
)
from ctdam.proc.modules.geomar_wildedit import WildeditGEOMAR
from ctdam.proc.modules.seabird_functions import (
    AlignCTD,
    BinAvg,
    CellTM,
    LoopRemoval,
    WFilter,
)
from ctdam.proc.utils import default_seabird_exe_path

mapper = {
    "airpressure": AirPressureCorrection(),
    "alignctd": AlignCTD(),
    "binavg": BinAvg(),
    "celltm": CellTM(),
    "create_bottle_file": create_bottle_file,
    "loop_removal": LoopRemoval(),
    "wfilter": WFilter(),
    "wildedit_geomar": WildeditGEOMAR(),
}

processing_functions = ExternalFunctions([gsw, sbs_proc])


def map_proc_name_to_class(module: str) -> Module:
    """
    Sets and maps the known processing modules to their respective
    module classes.

    Parameters
    ----------
    module: str :
        Name of the module, that is being used inside the config.

    Returns
    -------

    """
    if module in processing_functions.list_of_function_names():
        return ExternalFunctionCaller(module, processing_functions)
    else:
        return mapper[module.lower()]


def get_list_of_custom_exes(
    path_to_custom_exe_dir: Path | str | None = None,
) -> list[str]:
    if isinstance(path_to_custom_exe_dir, Path | str):
        return [exe.stem for exe in Path(path_to_custom_exe_dir).glob("*.exe")]
    else:
        return []


def get_list_of_installed_seabird_modules() -> list[str]:
    seabird_path = default_seabird_exe_path()
    return [str(file.stem)[:-1] for file in seabird_path.glob("*W.exe")]


def get_dict_of_available_processing_modules(
    path_to_custom_exe_dir: Path | str | None = None,
) -> dict:
    proc_dict = {
        "custom": [
            *list(mapper.values()),
            *get_list_of_custom_exes(path_to_custom_exe_dir),
        ],
        "seabird_exes": get_list_of_installed_seabird_modules(),
        **processing_functions,
    }
    return proc_dict



import importlib.metadata
import logging
import shutil
import sys
from pathlib import Path

from ctdam import APPNAME
from ctdam.conv import decode_hex
from ctdam.parser import CTDData

try:
    import typer
    from platformdirs import user_config_dir, user_log_dir
    from typing_extensions import Annotated
except (ImportError, ModuleNotFoundError, TypeError):
    sys.exit(
        "The 'cli' extra is required to use this feature. Install with: uv pip install ctdam[cli]"
    )

from ctdam.parser import HexCollection
from ctdam.proc.procedure import Procedure
from ctdam.proc.settings import Configuration
from ctdam.proc.utils import default_seabird_exe_path

logger = logging.getLogger(__name__)


log_file_path = (
    Path(user_log_dir(APPNAME)).joinpath(APPNAME).with_suffix(".log")
)
config_dir = Path(user_config_dir(APPNAME))
VIS_CONFIG_NAME = "vis_config.toml"
app = typer.Typer()


@app.callback()
def common(
    ctx: typer.Context,
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Enable verbose output."
    ),
):
    ctx.obj = {"verbose": verbose}
    if not log_file_path.exists():
        log_file_path.parent.mkdir(parents=True, exist_ok=True)
    logging.basicConfig(
        level=logging.DEBUG if verbose else logging.WARNING,
        format="%(asctime)s - [%(levelname)s] - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[
            logging.FileHandler(log_file_path),
            logging.StreamHandler(),
        ],
    )


@app.command()
def run(
    processing_target: Annotated[
        str,
        typer.Argument(
            help="The target file to process.",
        ),
    ],
    path_to_configuration: Annotated[
        str,
        typer.Argument(
            help="The path to the configuration file.",
        ),
    ],
    procedure_fingerprint_directory: Annotated[
        str,
        typer.Option(
            "--fingerprint",
            "-f",
            help="The path to a fingerprint directory. If none given, no fingerprints will be created.",
        ),
    ] = "",
    file_type_dir: Annotated[
        str,
        typer.Option(
            "--file-type",
            "-t",
            help="The path to a file type directory. If none given, the files will not be separated into file type directories.",
        ),
    ] = "",
    plot: Annotated[
        bool,
        typer.Option(
            "--plot",
            "-p",
            help="Whether to plot the file after processing.",
        ),
    ] = False,
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose",
            "-v",
            help="An option to allow more verbose output.",
        ),
    ] = False,
):
    """
    Processes one target file using the given procedure workflow file.
    """
    path_to_config = Path(path_to_configuration)
    if path_to_config.exists():
        config = Configuration(path_to_config)
    else:
        sys.exit("Could not find the configuration file.")
    config["input"] = processing_target
    Procedure(
        configuration=config,
        procedure_fingerprint_directory=procedure_fingerprint_directory,
        file_type_dir=file_type_dir,
        plot=plot,
        verbose=verbose,
    )


@app.command()
def convert(
    input_dir: Annotated[
        str,
        typer.Argument(
            help="The data directory with the target .hex files.",
        ),
    ],
    psa_path: Annotated[
        str,
        typer.Option(
            "--psa_path",
            "-s",
            help="The path to the .psa for datcnv. If none given, hex2py will be used for conversion.",
        ),
    ] = "",
    output_dir: Annotated[
        str,
        typer.Option(
            "--output",
            "-o",
            help="The directory to store the converted .cnv files in.",
        ),
    ] = "",
    xmlcon_dir: Annotated[
        str,
        typer.Option(
            "--xmlcons",
            "-x",
            help="The directory to look for .xmlcon files.",
        ),
    ] = "",
    pattern: Annotated[
        str,
        typer.Option(
            "--pattern",
            "-p",
            help="A name pattern to filter the target .hex files with.",
        ),
    ] = "",
) -> list[Path]:
    """
    Converts a list of Sea-Bird raw data files (.hex) to .cnv files.
    Does either use an explicit list of paths or searches for all .hex files in
    the given directory.
    """
    if not output_dir:
        output_dir = input_dir
    if not xmlcon_dir:
        xmlcon_dir = input_dir
    hexes = HexCollection(
        path_to_files=Path(input_dir),
        pattern=pattern,
        file_suffix="hex",
        path_to_xmlcons=Path(xmlcon_dir),
    )
    resulting_cnvs = []
    if psa_path:
        proc_config = {
            "output_dir": output_dir,
            "modules": {
                "datcnv": {"psa": psa_path},
            },
        }
        procedure = Procedure(
            proc_config,
            auto_run=False,
            verbose=True if len(hexes) == 1 else False,
        )
    with typer.progressbar(hexes, label="Converting files:") as progress:
        for hex in progress:
            try:
                if psa_path:
                    result = procedure.run(hex.path_to_file)
                else:
                    result = decode_hex(hex)
                    file_name = result.metadata_source.file_name
                    result.to_cnv(
                        Path(output_dir)
                        .joinpath(file_name)
                        .with_suffix(".cnv")
                    )
            except Exception as e:
                logger.error(f"Failed to convert: {hex.path_to_file}, {e}")
            else:
                resulting_cnvs.append(result)

    return resulting_cnvs


@app.command()
def batch(
    input_dir: Annotated[
        str,
        typer.Argument(
            help="The data directory with the target files.",
        ),
    ],
    config: Annotated[
        str,
        typer.Argument(
            help="Either an explicit config as dict or a path to a .toml config file.",
        ),
    ],
    pattern: Annotated[
        str,
        typer.Option(
            "--pattern",
            "-p",
            help="A name pattern to filter the target files with.",
        ),
    ] = ".hex",
) -> list[Path] | list[CTDData]:
    """
    Applies a processing config to multiple .hex or. cnv files.
    """
    resulting_cnvs = []
    if isinstance(config, dict):
        proc_config = config
    else:
        proc_config = Configuration(config)
    procedure = Procedure(proc_config, auto_run=False)
    with typer.progressbar(
        Path(input_dir).rglob(f"*{pattern}*"), label="Processing files:"
    ) as progress:
        for file in progress:
            try:
                result = procedure.run(file)
            except Exception as e:
                logger.error(f"Error when processing {file}: {e}")
            else:
                resulting_cnvs.append(result)
    return resulting_cnvs


try:
    from processing.gui.procedure_config_view import run_gui
except ImportError:
    pass
else:

    @app.command()
    def edit(file: str):
        """
        Opens a procedure workflow file in GUI for editing.
        """
        run_gui(file)


@app.command()
def show(file: typer.FileText):
    """
    Display the contents of a procedure workflow file.
    """
    content = file.read()
    print(content, end="")


try:
    from processing.visualize import basic_bokeh_plot, cruise_plots
except ImportError:
    pass
else:

    @app.command()
    def plot(
        cnv: Annotated[
            str,
            typer.Argument(
                help="The path to the cnv file.",
            ),
        ] = "",
        output_directory: Annotated[
            str,
            typer.Option(
                "--output_directory",
                "-d",
                help="The path to the output_directory.",
            ),
        ] = "html",
        output_name: Annotated[
            str,
            typer.Option(
                "--output_name",
                "-o",
                help="The name of the output .html .",
            ),
        ] = "",
        save: Annotated[
            bool,
            typer.Option(
                "--save",
                "-s",
                help="Whether to save the plot as a .html file.",
            ),
        ] = False,
        metadata: Annotated[
            bool,
            typer.Option(
                "--metadata",
                "-m",
                help="Whether to display .cnv file metadata in the plot.",
            ),
        ] = False,
    ):
        """
        Plot a cnv file.
        """
        if output_name:
            save = True
        basic_bokeh_plot(
            ctd_data=cnv,
            output_name=str(output_name),
            output_directory=output_directory,
            print_plot=save,
            metadata=metadata,
        )

    @app.command()
    def vis(
        directory: Annotated[
            str,
            typer.Argument(
                help="The path to the target directory holding the .cnv files.",
            ),
        ] = "",
        output_directory: Annotated[
            str,
            typer.Option(
                "--output_directory",
                "-d",
                help="The path to the output directory.",
            ),
        ] = "html",
        output_name: Annotated[
            str,
            typer.Option(
                "--output_name",
                "-o",
                help="The name of the output .html file.",
            ),
        ] = "main.html",
        embed_contents: Annotated[
            bool,
            typer.Option(
                "--embed_contents",
                "-e",
                help="Whether to embed the target .html files or just link to them.",
            ),
        ] = False,
        html_title: Annotated[
            str,
            typer.Option(
                "--html_title",
                "-t",
                help="The title that will be used inside the .html file.",
            ),
        ] = "",
        overwrite: Annotated[
            bool,
            typer.Option(
                "--overwrite",
                "-w",
                help="Whether to overwrite existing plot .html files.",
            ),
        ] = False,
        no_new_plots: Annotated[
            bool,
            typer.Option(
                "--no_new_plots",
                "-p",
                help="Whether no new plot .html files should be created.",
            ),
        ] = False,
        size_limit: Annotated[
            int,
            typer.Option(
                "--size_limit",
                "-l",
                help="""File size limit in MB to which plots will be created.
             Very large files can slow down the visualizer considerabily.""",
            ),
        ] = 10,
        filter: Annotated[
            str,
            typer.Option(
                "--filter",
                "-f",
                help="The files to select for visualization.",
            ),
        ] = "",
        file_type: Annotated[
            str,
            typer.Option(
                "--file_type",
                "-y",
                help="The data type to plot.",
            ),
        ] = "cnv",
    ):
        """
        Create a main html that incorporates the individual .html plots.
        """

        _check_config_path()

        output_path = cruise_plots(
            directory=directory,
            output_directory=output_directory,
            output_name=output_name,
            embed_contents=embed_contents,
            html_title=html_title,
            overwrite=overwrite,
            no_new_plots=no_new_plots,
            size_limit=size_limit,
            filter=filter,
            config_path=config_dir.joinpath(VIS_CONFIG_NAME),
            file_type=file_type,
        )
        print(f"Created main .html file: {output_path}")


def _check_config_path():
    if not config_dir.exists():
        config_dir.mkdir(parents=True)
    vis_config_path = config_dir.joinpath(VIS_CONFIG_NAME)
    if not vis_config_path.exists():
        shutil.copy(
            Path(__file__).parent.joinpath(VIS_CONFIG_NAME), vis_config_path
        )


@app.command()
def check():
    """
    Assures that all requirements to use this tool are met.
    """
    if not default_seabird_exe_path().exists():
        print(
            "You are missing a Sea-Bird Processing installation or are not using the default path. Please ensure that a valid installation can be found in Program Files (x86)/Sea-Bird/SBEDataProcessing-Win32/"
        )
    else:
        print("All set, you are ready to go.")
    try:
        from processing.gui.procedure_config_view import run_gui  # noqa: F401
    except ImportError:
        print(
            "\nIf you want to use a GUI to edit your ctd processing workflows, install the additional dependencies via 'uvx--from ctdam[gui] ctdam'"
        )
    try:
        from processing.visualize import (  # noqa: F401
            basic_bokeh_plot,
            cruise_plots,
        )
    except ImportError:
        print(
            "\nIf you want to use the plotting capabilities, install the additional dependencies via 'uvx--from ctdam[vis] ctdam'"
        )


@app.command()
def log(
    number_of_entries: Annotated[
        int, typer.Argument(help="The number of entries to print.")
    ] = 30,
):
    """
    Prints the last x entries of the log file.
    """
    if not log_file_path.exists():
        return
    lines = log_file_path.read_text().splitlines()
    last_x_lines = lines[-number_of_entries:]
    for line in last_x_lines:
        print(line)


@app.command()
def version():
    """
    Displays the version number of this software.
    """
    print(importlib.metadata.version("ctdam"))


if __name__ == "__main__":
    app()



import logging
from collections import UserDict
from pathlib import Path

from tomlkit import dumps, table
from tomlkit.toml_file import TOMLFile

logger = logging.getLogger(__name__)


class Configuration(UserDict):
    """
    The internal representation of the .toml configuration files, that store
    the ctdam.proc information.

    Allows the interaction with these config files to be pretty much equivalent
    to a basic python dictionary.
    """

    def __init__(self, path: Path | str, data: dict | None = None):
        self.path = Path(path)
        if data is None:
            self.data = TOMLFile(path).read()
        else:
            # toml cannot handle Path objects
            for key, value in data.items():
                if isinstance(value, Path):
                    data[key] = str(value)
            self.data = data

    def __str__(self):
        return str(self.path)

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, keys, value):
        self.modify([keys], value)

    def write(self, path_to_write=None):
        """
        Writes the ctdam.proc information to a file.

        Parameters
        ----------
        path_to_write :
            The path to write the file to. If none given, the input path will
            be used.
             (Default value = None)
        """
        output_path = self.path
        if path_to_write:
            output_path = path_to_write
        try:
            with open(output_path, "w") as file:
                file.write(dumps(self.data))
        except IOError as error:
            logger.error(f"Could not write configuration file: {error}")

    def modify(self, key, value):
        """
        Allows the access and modification of nested data points.

        Parameters
        ----------
        key :

        value :


        Returns
        -------

        """
        # TODO: allow addition to a specified position
        try:
            if isinstance(key, list):
                current_section = self.data
                for position in key[:-1]:
                    current_section = current_section.get(position, table())

                current_section[key[-1]] = value
            else:
                self.data.update({key: value})
        except ValueError as error:
            logger.error(f"Value modification failed: {error}")


class IncompleteConfigFile(Exception):
    """An exception to indicate misformed configuration files."""

    def __init__(self, message):
        super().__init__(message)
